from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, List, Union

import numpy as np
from omegaconf import MISSING

from ertk.config import ERTKConfig
from ertk.preprocessing.base import AudioClipProcessor, FeatureExtractor
from ertk.preprocessing.spectrogram import spectrogram
from ertk.tensorflow.utils import init_gpu_memory_growth
from ertk.utils import frame_array, is_mono_audio


def _spectrogram(x: np.ndarray, sr: float) -> np.ndarray:
    return spectrogram(
        x,
        sr,
        kind="mel",
        pre_emphasis=0,
        window_size=0.025,
        window_shift=0.01,
        n_mels=64,
        fmin=125,
        fmax=7500,
    )


@dataclass
class YamnetExtractorConfig(ERTKConfig):
    model_dir: str = MISSING


class YamnetExtractor(
    FeatureExtractor, AudioClipProcessor, fname="yamnet", config=YamnetExtractorConfig
):
    config: YamnetExtractorConfig

    def __init__(self, config: YamnetExtractorConfig) -> None:
        super().__init__(config)

        import tensorflow as tf

        from ._yamnet.params import Params
        from ._yamnet.yamnet import yamnet

        init_gpu_memory_growth()

        input_layer = tf.keras.Input((96, 64))
        predictions, embeddings = yamnet(input_layer, Params())
        model = tf.keras.Model(inputs=input_layer, outputs=[predictions, embeddings])
        model.load_weights(str(Path(config.model_dir, "yamnet.h5")))
        model(tf.zeros((1, 96, 64)))
        self.model = model

    def process_instance(self, x: np.ndarray, **kwargs) -> np.ndarray:
        if is_mono_audio(x):
            x = _spectrogram(x, kwargs.pop("sr"))
        frames = frame_array(x, 96, 48, pad=True, axis=0)
        _, embeddings = self.model(frames)
        return np.mean(embeddings, 0)

    def process_batch(
        self, batch: Union[Iterable[np.ndarray], np.ndarray], **kwargs
    ) -> List[np.ndarray]:
        batch = list(batch)
        if all(is_mono_audio(x) for x in batch):
            sr = kwargs.pop("sr")
            batch = [_spectrogram(x, sr) for x in batch]
        frames = [frame_array(spec, 96, 48, pad=True, axis=0) for spec in batch]
        slices = np.cumsum([len(x) for x in frames])[:-1]
        frames = np.concatenate(frames)
        frames *= np.log(10.0) / 20  # Rescale dB power to ln(abs(X))

        _, embeddings = self.model(frames)
        return [np.mean(x, 0) for x in np.split(embeddings, slices)]

    @property
    def dim(self) -> int:
        return 1024

    @property
    def is_sequence(self) -> bool:
        return False


@dataclass
class VGGishExtractorConfig(ERTKConfig):
    model_dir: str = MISSING


class VGGishExtractor(
    FeatureExtractor, AudioClipProcessor, fname="vggish", config=VGGishExtractorConfig
):
    config: VGGishExtractorConfig

    def __init__(self, config: VGGishExtractorConfig) -> None:
        super().__init__(config)

        import tensorflow.compat.v1 as tf

        from ._vggish.vggish_postprocess import Postprocessor
        from ._vggish.vggish_slim import define_vggish_slim, load_vggish_slim_checkpoint

        tf.disable_eager_execution()
        for gpu in tf.config.list_physical_devices("GPU"):
            tf.config.experimental.set_memory_growth(gpu, True)

        pca_params = str(Path(config.model_dir, "vggish_pca_params.npz"))
        self.processor = Postprocessor(pca_params)

        self.sess = tf.Session()

        define_vggish_slim()
        ckpt = str(Path(config.model_dir, "vggish_model.ckpt"))
        load_vggish_slim_checkpoint(self.sess, ckpt)

        self.features_tensor = self.sess.graph.get_tensor_by_name(
            "vggish/input_features:0"
        )
        self.embedding_tensor = self.sess.graph.get_tensor_by_name("vggish/embedding:0")

    def __del__(self):
        self.sess.close()

    def _process_frames(self, frames: np.ndarray) -> np.ndarray:
        [embeddings] = self.sess.run(
            [self.embedding_tensor], feed_dict={self.features_tensor: frames}
        )
        return self.processor.postprocess(embeddings)

    def process_instance(self, x: np.ndarray, **kwargs) -> np.ndarray:
        if is_mono_audio(x):
            x = _spectrogram(x, kwargs.pop("sr"))
        frames = frame_array(x, 96, 96, pad=True, axis=0)
        return np.mean(self._process_frames(frames), 0)

    def process_batch(
        self, batch: Union[Iterable[np.ndarray], np.ndarray], **kwargs
    ) -> List[np.ndarray]:
        batch = list(batch)
        if all(is_mono_audio(x) for x in batch):
            sr = kwargs.pop("sr")
            batch = [_spectrogram(x, sr) for x in batch]
        frames = [frame_array(spec, 96, 96, pad=True, axis=0) for spec in batch]
        slices = np.cumsum([len(x) for x in frames])[:-1]
        frames = np.concatenate(frames)
        frames *= np.log(10.0) / 20  # Rescale dB power to ln(abs(X))
        return [np.mean(x, 0) for x in np.split(self._process_frames(frames), slices)]

    @property
    def dim(self) -> int:
        return 128

    @property
    def is_sequence(self) -> bool:
        return False



import warnings
from dataclasses import dataclass
from typing import Optional

import librosa
import numpy as np

from ertk.config import ERTKConfig

from .base import AudioClipProcessor, FeatureExtractor


def spectrogram(
    audio: np.ndarray,
    sr: float,
    kind: str = "mel",
    pre_emphasis: float = 0,
    window_size: float = 0.025,
    window_shift: float = 0.01,
    n_fft: Optional[int] = None,
    n_mels: int = 128,
    htk_mel: bool = False,
    n_chroma: int = 12,
    clip_db: Optional[float] = None,
    fmin: float = 0,
    fmax: Optional[float] = 8000,
    power: int = 2,
    to_db: bool = True,
):
    """General purpose spectrogram pipeline. Calculates spectrogram with
    optional pre-emphasis, clipping, mel/chroma transform, power and
    conversion to dB.

    Parameters
    ----------
    audio: np.ndarray
        Audio data. This should be a 1D array, or a 2D array with shape
        (1, n_samples).
    sr: int
        Sample rate.
    kind: str
        The kind of spectrogram to calculate. "mel" calculates a mel
        spectrogram, whereas "stft" just uses linear frequency.
    pre_emphasis: float
        Amount of pre-emphasis to apply. Default is 0 (no pre-emphasis).
    window_size: float
        Size of window in seconds. Default is 0.025 (25 ms).
    window_shift: float
        Amount window moves each frame, in seconds. Default is 0.01 (10
        ms).
    n_fft: int, optional
        Number of FFT bins. Default is to use the next power of 2
        greater than `window_size * sr`.
    n_mels: int
        Number of mel bands.
    htk_mel: bool,
        Use HTK formula for mel calculation.
    n_chroma: int
        Number of chroma bands.
    clip_db: float, optional
        Whether to clip noise floor at a given level in dB below
        maximum. Default is `None` which does no clipping. This is only
        used if `to_db` is `True`.
    fmin: float
        Minimum frequency for mel bands. Default is 0.
    fmax: float, optional
        Maximum frequency for mel bands. Default is 8000.
    power: int
        Raise spectrogram magnitudes to this power. Default is 2 which
        is the usual power spectrogram.
    to_db: bool
        Whether to convert spectrogram to dB units. Default is `True`.
        Note that this is mainly only useful if `power == 2`.

    Returns
    -------
    melspec: np.ndarray
        An array of shape (n_frames, n_mels) containing mel spectrogram.
        If power=2 this is the power spectrogram, and if to_db is `True`
        then this is in dB units.
    """
    audio = np.squeeze(audio)
    if audio.ndim != 1:
        raise ValueError(f"Audio should be in mono format, got shape {audio.shape}")

    window_samples = int(window_size * sr)
    stride_samples = int(window_shift * sr)

    # Pre-emphasis
    if pre_emphasis > 0:
        audio = librosa.effects.preemphasis(audio, pre_emphasis)

    # Spectrogram
    if not n_fft:
        n_fft = 2 ** int(np.ceil(np.log2(window_samples)))
    warnings.simplefilter("ignore", UserWarning)
    spec = librosa.stft(
        audio,
        n_fft=n_fft,
        hop_length=stride_samples,
        win_length=window_samples,
        center=True,
        window="hann",
        pad_mode="reflect",
    )
    spec = np.abs(spec) ** power
    if kind == "mel":
        spec = librosa.feature.melspectrogram(
            S=spec, n_mels=n_mels, fmin=fmin, fmax=fmax, htk=htk_mel
        )
    elif kind == "chroma":
        spec = librosa.feature.chroma_stft(S=spec, n_chroma=n_chroma)
    warnings.simplefilter("default", UserWarning)

    # Optional dB calculation
    if to_db:
        spec = librosa.power_to_db(spec, ref=np.max, top_db=clip_db)

    return spec.T


@dataclass
class SpectrogramExtractorConfig(ERTKConfig):
    kind: str = "mel"
    pre_emphasis: float = 0
    window_size: float = 0.025
    window_shift: float = 0.01
    n_mels: int = 128
    htk_mel: bool = False
    n_fft: Optional[int] = None
    n_chroma: int = 12
    clip_db: Optional[float] = None
    fmin: float = 0
    fmax: Optional[float] = 8000
    power: int = 2
    to_db: bool = True


class SpectrogramExtractor(
    FeatureExtractor,
    AudioClipProcessor,
    fname="spectrogram",
    config=SpectrogramExtractorConfig,
):
    config: SpectrogramExtractorConfig

    def process_instance(self, x: np.ndarray, **kwargs) -> np.ndarray:
        return spectrogram(x, sr=kwargs.pop("sr"), **self.config)  # type: ignore

    @property
    def dim(self) -> int:
        return self.config.n_mels

    @property
    def is_sequence(self) -> bool:
        return True



import joblib
import tqdm


# Class adapted from user394430's answer here:
# https://stackoverflow.com/a/61900501/10044861
# Licensed under CC BY-SA 4.0
class TqdmParallel(joblib.Parallel):
    """Convenience class that acts identically to joblib.Parallel except
    it uses a tqdm progress bar.

    Parameters
    ----------
    total: int
        Total number of items in the iterable.
    desc: str
        Progress bar description.
    unit: str
        Progress bar unit.
    leave: bool
        Whether to leave the progress bar after completion.
    **kwargs: dict
        Other keyword args passed to joblib.Parallel.
    """

    def __init__(
        self,
        total: int = 1,
        desc: str = "",
        unit: str = "it",
        leave: bool = True,
        **kwargs,
    ):
        self.tqdm_args = {
            "total": total,
            "desc": desc,
            "unit": unit,
            "leave": leave,
            "disable": None,
        }
        kwargs["verbose"] = 0
        super().__init__(**kwargs)

    def __call__(self, iterable):
        with tqdm.tqdm(**self.tqdm_args) as self.pbar:
            return super().__call__(iterable)

    def print_progress(self):
        self.pbar.n = self.n_completed_tasks
        self.pbar.refresh()



from pathlib import Path

import click

from ertk.dataset import read_features, write_features


@click.command()
@click.argument("input", type=click.Path(exists=True, dir_okay=False, path_type=Path))
@click.argument("output", type=Path, required=False)
@click.option(
    "--names",
    "names_file",
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
)
def main(input: Path, output: Path, names_file: Path):
    """Remove instances from INPUT features that aren't in the NAMES
    file, and write to OUTPUT. If OUTPUT is not given, overwrites INPUT
    in-place.
    """
    with open(names_file) as fid:
        keep_names = {Path(x.strip()).stem for x in fid}
    data = read_features(input)
    idx, names = zip(*[(i, n) for i, n in enumerate(data.names) if n in keep_names])

    output = output or input
    write_features(
        output,
        corpus=data.corpus,
        names=list(names),
        features=data.features[idx],
        feature_names=data.feature_names,
    )
    print(f"Wrote features to {input}")


if __name__ == "__main__":
    main()



"""Implementation of the model from [1].

[1] Z. Zhao et al., "Exploring Deep Spectrum Representations via
Attention-Based Recurrent and Convolutional Neural Networks for Speech
Emotion Recognition," IEEE Access, vol. 7, pp. 97515–97525, 2019, doi:
10.1109/ACCESS.2019.2928625.
"""

import tensorflow as tf
from tensorflow.keras.layers import (
    LSTM,
    RNN,
    BatchNormalization,
    Bidirectional,
    Conv2D,
    Dense,
    Dropout,
    Input,
    LSTMCell,
    MaxPool2D,
    ReLU,
    Reshape,
    concatenate,
)
from tensorflow.keras.models import Model

from .layers import Attention1D

__all__ = ["model"]


def model(n_features: int, n_classes: int, steps: int = 512):
    inputs = Input((steps, n_features))

    # BLSTM
    # blstm_seq = Bidirectional(RNN([LSTMCell(128, dropout=0.2), LSTMCell(128)],
    #                               return_sequences=True))(inputs)
    blstm_seq = Bidirectional(LSTM(128, dropout=0.2, return_sequences=True))(inputs)
    seq_att = Attention1D()(blstm_seq)

    # FCN
    expanded = tf.expand_dims(inputs, -1)
    conv1 = Conv2D(64, 8)(expanded)
    conv1 = BatchNormalization()(conv1)
    conv1 = ReLU()(conv1)
    conv1 = MaxPool2D(2)(conv1)
    conv1 = Dropout(0.2)(conv1)

    conv2 = Conv2D(128, 5)(conv1)
    conv2 = BatchNormalization()(conv2)
    conv2 = ReLU()(conv2)
    conv2 = MaxPool2D(2)(conv2)
    conv2 = Dropout(0.2)(conv2)

    conv3 = Conv2D(128, 3)(conv2)
    conv3 = BatchNormalization()(conv3)
    conv3 = ReLU()(conv3)
    conv3 = MaxPool2D(2)(conv3)
    conv3 = Dropout(0.2)(conv3)

    pool = Reshape((-1, 128))(conv3)
    pool = Dense(128, activation="tanh")(pool)
    fcn_att = Attention1D()(pool)

    concat = concatenate([seq_att, fcn_att])
    x = Dense(128, activation="relu")(concat)
    x = Dense(n_classes, activation="softmax")(x)
    return Model(inputs=inputs, outputs=x)



import csv
import typing
import warnings
from collections import Counter
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, List, Optional, Union

import arff
import librosa
import netCDF4
import numpy as np
import pandas as pd
import soundfile
from joblib import delayed

from ertk.dataset.utils import get_audio_paths, write_filelist
from ertk.utils import (
    PathOrStr,
    TqdmParallel,
    filter_kwargs,
    flat_to_inst,
    inst_to_flat,
    make_array_array,
)


def read_arff(path: PathOrStr, label: bool = False):
    path = Path(path)
    with open(path) as fid:
        data = arff.load(fid)

    attr_names = [x[0] for x in data["attributes"]]
    counts = Counter([x[0] for x in data["data"]])
    idx = slice(1, -1) if label else slice(1, None)

    return FeaturesData(
        corpus=data["relation"],
        names=list(counts.keys()),
        features=np.array([x[idx] for x in data["data"]]),
        slices=np.array(list(counts.values())),
        feature_names=attr_names[idx],
    )


def read_csv(path: PathOrStr, header: bool = True, label: bool = False):
    df = pd.read_csv(path, header=0 if header else None, converters={0: str})
    counts: typing.Counter[str] = Counter(df.iloc[:, 0])
    idx = slice(1, -1) if label else slice(1, None)

    return FeaturesData(
        names=list(counts.keys()),
        features=np.array(df.iloc[:, idx]),
        slices=np.array(list(counts.values())),
        feature_names=list(df.columns[idx]),
    )


def read_netcdf(path: PathOrStr):
    with netCDF4.Dataset(path) as dataset:
        return FeaturesData(
            corpus=dataset.corpus,
            names=list(dataset.variables["name"]),
            features=np.array(dataset.variables["features"], copy=False),
            slices=np.array(dataset.variables["slices"]),
            feature_names=list(dataset.variables["feature_names"]),
        )


def read_raw(path: PathOrStr, sample_rate: int = 16000):
    path = Path(path)
    filepaths = get_audio_paths(path)
    _audio = []

    def read_one_file(filepath):
        with warnings.catch_warnings():
            audio, _ = librosa.load(
                filepath, sr=sample_rate, mono=True, res_type="kaiser_fast"
            )
        return np.expand_dims(audio, -1)

    _audio = TqdmParallel(len(filepaths), desc="Reading audio", leave=False, n_jobs=-1)(
        delayed(read_one_file)(filepath) for filepath in filepaths
    )

    return FeaturesData(
        corpus=path.parent.stem,
        names=[x.stem for x in filepaths],
        features=np.concatenate(_audio),
        slices=[len(x) for x in _audio],
        feature_names=["pcm"],
    )


class FeaturesData:
    """Dataset features backend.

    Parameters
    ----------
    features: ndarray
        Features matrix. May be "flat" or per-instance. If flat,
        `slices` must also be present.
    names: list of str
        List of instance names.
    corpus: str
        Corpus for this dataset.
    feature_names: list of str, optional
        List of feature names. If not given, the features will me named
        feature1, feature2, etc.
    slices: list or ndarray:
        Slices corresponding to "flat" `features` matrix.
    dataset: DatasetBackend
        Other backend instance to copy attributes from. Takes priority
        over other passed args.
    """

    def __init__(
        self,
        features: Union[List[np.ndarray], np.ndarray],
        names: List[str],
        corpus: str = "",
        feature_names: Optional[List[str]] = None,
        slices: Union[np.ndarray, List[int], None] = None,
    ):
        self._corpus = corpus
        self._names = list(names)
        if isinstance(features, list):
            features = make_array_array(features)
        self._features = features
        if feature_names is None:
            feature_names = []
        self._feature_names = list(feature_names)

        if slices is not None:
            self._slices = np.array(slices)
            self._flat = features
            self._features = flat_to_inst(self.flat, slices)
        else:
            self._slices = np.ones(len(features), dtype=int)
            if len(features.shape) != 2 and features[0].ndim > 1:
                self._slices = np.array([len(x) for x in features])

        n_features = self.features[0].shape[-1]
        if len(self.feature_names) == 0:
            self._feature_names = [f"feature{i + 1}" for i in range(n_features)]
        elif len(self.feature_names) != n_features:
            raise ValueError(
                f"feature_names has {len(self.feature_names)} entries but there "
                f"are {n_features} features"
            )

    def make_contiguous(self):
        """Makes `self.features` a contiguous array that represents
        views of `self.flat`.
        """
        self._flat, self._slices = inst_to_flat(self.features)
        self._features = flat_to_inst(self.flat, self.slices)

    def write(self, path: PathOrStr, **kwargs):
        """Write features to any file format."""

        path = Path(path)
        path.parent.mkdir(exist_ok=True, parents=True)
        meth = FeaturesData._write_backends.get(path.suffix)
        if meth is None:
            raise ValueError(f"File format {path.suffix} not supported.")
        meth(self, path, **filter_kwargs(kwargs, meth))

    def write_arff(self, path: PathOrStr):
        data: Dict[str, Any] = {"relation": self.corpus}
        data["attributes"] = [("name", "STRING")] + [
            (x, "NUMERIC") for x in self.feature_names
        ]
        data["data"] = []
        # The liac-arff library can only dump all at once
        self.make_contiguous()
        for name, row in zip(np.repeat(self.names, self.slices), self.flat):
            data["data"].append([name] + list(row))

        with open(path, "w") as fid:
            arff.dump(data, fid)

    def write_csv(self, path, header: bool = True):
        with open(path, "w") as fid:
            writer = csv.writer(fid)
            if header:
                writer.writerow(["name"] + self.feature_names)
            for name, inst in zip(self.names, self.features):
                if len(inst.shape) == 2:
                    for row in inst:
                        writer.writerow([name] + list(row))
                else:
                    writer.writerow([name] + list(inst))

    def write_netcdf(self, path: PathOrStr):
        dataset = netCDF4.Dataset(path, "w")
        dataset.createDimension("instance", len(self.names))
        dataset.createDimension("concat", sum(self.slices))
        dataset.createDimension("features", self.features[0].shape[-1])

        _slices = dataset.createVariable("slices", int, ("instance",))
        _slices[:] = np.array(self.slices)

        filename = dataset.createVariable("name", str, ("instance",))
        filename[:] = np.array(self.names)

        _features = dataset.createVariable(
            "features", np.float32, ("concat", "features")
        )
        idx = np.cumsum(self.slices)
        for i, x in enumerate(self.features):
            end = idx[i]
            start = end - self.slices[i]
            _features[start:end, :] = x

        _feature_names = dataset.createVariable("feature_names", str, ("features",))
        _feature_names[:] = np.array(self.feature_names)

        dataset.setncattr_string("corpus", self.corpus)
        dataset.close()

    def write_raw(self, path: PathOrStr, sr: int = 16000):
        output_dir = Path(path).with_suffix("")
        output_dir.mkdir(exist_ok=True, parents=True)
        for name, audio in zip(self.names, self.features):
            output_path = output_dir / f"{name}.wav"
            soundfile.write(output_path, audio, subtype="PCM_16", samplerate=sr)
        write_filelist(output_dir.glob("*.wav"), path)

    @property
    def features(self) -> np.ndarray:
        """Instance features matrix. This may be 2D/3D contiguous or
        variable length 3D.
        """
        return self._features

    @property
    def flat(self) -> np.ndarray:
        """2D "flat" features matrix. This has shape (N, F), where
        N = sum(slices), F is the number of features.
        """
        return self._flat

    @property
    def slices(self) -> np.ndarray:
        """Slices corresponding to flat features matrix. It should be
        the case that `features = flat_to_inst(flat, slices)` and
        `flat, slices = inst_to_flat(features)`.
        """
        return self._slices

    @property
    def names(self) -> List[str]:
        """Instance names."""
        return self._names

    @property
    def feature_names(self) -> List[str]:
        """Names of features in feature matrix."""
        return self._feature_names

    @property
    def corpus(self) -> str:
        """Corpus ID."""
        return self._corpus

    def __len__(self) -> int:
        return len(self.names)

    _write_backends: Dict[str, Callable] = {
        ".arff": write_arff,
        ".csv": write_csv,
        ".nc": write_netcdf,
        ".txt": write_raw,
    }


_READ_BACKENDS: Dict[str, Callable[..., FeaturesData]] = {
    ".nc": read_netcdf,
    ".arff": read_arff,
    ".csv": read_csv,
    ".txt": read_raw,
}


def register_format(
    suffix: str,
    read: Optional[Callable[..., FeaturesData]],
    write: Optional[Callable[..., None]],
):
    """Register new read/write functions for a given file format."""

    if suffix[0] != ".":
        suffix = "." + suffix
    if read is not None:
        _READ_BACKENDS[suffix] = read
    if write is not None:
        FeaturesData._write_backends[suffix] = write


def find_features_file(files: Iterable[Path]) -> Path:
    files = list(files)
    for suffix in _READ_BACKENDS:
        file = next((x for x in files if x.suffix == suffix), None)
        if file is not None:
            return file
    raise FileNotFoundError(f"No features found in {files}")


def read_features(path: PathOrStr, **kwargs) -> FeaturesData:
    """Read features from given path."""

    path = Path(path)
    meth = _READ_BACKENDS.get(path.suffix)
    if meth is None:
        raise ValueError(f"File format {path.suffix} not supported.")
    return meth(path, **filter_kwargs(kwargs, meth))


def write_features(
    path: PathOrStr,
    features: Union[List[np.ndarray], np.ndarray],
    names: List[str],
    corpus: str = "",
    feature_names: List[str] = None,
    slices: Union[np.ndarray, List[int]] = None,
    **kwargs,
):
    """Convenience function to write features to given path. Arguments
    are the same as those passed to FeaturesData.
    """

    FeaturesData(
        features, names, corpus=corpus, slices=slices, feature_names=feature_names
    ).write(path, **kwargs)



import logging
import subprocess
from pathlib import Path
from typing import Iterable, List, Mapping

from joblib import delayed

from ertk.utils import PathOrStr, TqdmParallel

# The below is because mypy does not yet support type narrowing upon redefinition
# mypy: allow-redefinition


def get_audio_paths(path: PathOrStr, absolute: bool = True) -> List[Path]:
    """Given a path to a dir or list of audio files, return a sequence
    of absolute paths to those files. Note that this method handles
    relative paths but returns absolute paths, and doesn't resolve
    canonical paths (i.e. it doesn't follow symlinks.)

    Parameters
    ----------
    file: pathlike or str
        Path to a directory containing audio clips, or a file containing
        a list of paths to audio clips.
    absolute: bool
        If `True`, return absolute paths, otherwise return the paths
        already in the file/directory. Default is `True`.

    Returns
    -------
    List of paths to audio files. Paths will be absolute paths if
    `absolute=True`.
    """
    path = Path(path)
    if path.is_dir():
        paths = [x for x in path.glob("*") if not x.is_dir()]
    else:
        with open(path) as fid:
            paths = [path.parent / x.strip() for x in fid]
    if absolute:
        return [x.absolute() for x in paths]
    return paths


def resample_rename_clips(mapping: Mapping[Path, Path]):
    """Resample given audio clips to 16 kHz 16-bit WAV.

    Parameters
    ----------
    mapping: mapping
        Mapping from source files to destination files.
    """
    dst_dirs = {x.parent for x in mapping.values()}
    for dir in dst_dirs:
        dir.mkdir(exist_ok=True, parents=True)

    opts = ["-nostdin", "-ar", "16000", "-sample_fmt", "s16", "-ac", "1", "-y"]
    logging.info(f"Resampling {len(mapping)} audio files")
    logging.info(f"Using FFmpeg options: {' '.join(opts)}")
    TqdmParallel(desc="Resampling audio", total=len(mapping), unit="file", n_jobs=-1)(
        delayed(subprocess.run)(
            ["ffmpeg", "-i", str(src), *opts, str(dst)],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.STDOUT,
        )
        for src, dst in mapping.items()
    )


def resample_audio(paths: Iterable[PathOrStr], dir: PathOrStr):
    """Resample given audio clips to 16 kHz 16-bit WAV, and place in
    direcotory given by `dir`.

    Parameters
    ----------
    paths: iterable of Path
        A collection of paths to audio files to resample.
    dir: Pathlike or str
        Output directory.
    """
    paths = list(map(Path, paths))
    if len(paths) == 0:
        raise FileNotFoundError("No audio files found.")

    resample_rename_clips({x: Path(dir, f"{x.stem}.wav") for x in paths})


def write_filelist(paths: Iterable[PathOrStr], path: PathOrStr):
    """Write sorted file list.

    Parameters
    ----------
    paths: iterable of Path
        Paths to audio clips.
    name: str
        Name of resulting file list.
    """
    paths = sorted(map(Path, paths), key=lambda p: p.stem)
    path = Path(path)
    if path.suffix == "":
        path = path.with_suffix(".txt")
    with open(path, "w") as fid:
        fid.write("\n".join(list(map(str, paths))) + "\n")



"""Implementation of model from [1].

[1] A. O. Iskhakova, D. A. Wolf, R. R. Galin, and M. V. Mamchenko, ‘1-D
convolutional neural network based on the inner ear principle to
automatically assess human’s emotional state’, E3S Web Conf., vol. 224,
p. 01023, 2020, doi: 10.1051/e3sconf/202022401023.
"""

from tensorflow.keras import Model
from tensorflow.keras.layers import (
    BatchNormalization,
    Conv1D,
    Dense,
    Dropout,
    Flatten,
    Input,
    MaxPool1D,
    ReLU,
    Reshape,
)

__all__ = ["model"]


def model(n_features: int, n_classes: int) -> Model:
    inputs = Input((n_features,))
    x = Reshape((n_features, 1))(inputs)

    # Layer 1
    x = Conv1D(128, 4, activation="relu", padding="same")(x)

    # Layer 2
    x = Conv1D(128, 4, padding="same")(x)
    x = BatchNormalization()(x)
    x = ReLU()(x)
    x = Dropout(0.25)(x)
    x = MaxPool1D(2)(x)

    # Layers 3-5
    x = Conv1D(64, 4, activation="relu", padding="same")(x)
    x = Conv1D(64, 4, activation="relu", padding="same")(x)
    x = Conv1D(64, 4, activation="relu", padding="same")(x)

    # Layer 6
    x = Conv1D(64, 4, padding="same")(x)
    x = BatchNormalization()(x)
    x = ReLU()(x)
    x = Dropout(0.25)(x)
    x = MaxPool1D(2)(x)

    # Layers 7-8
    x = Conv1D(32, 2, activation="relu", padding="same")(x)
    x = Conv1D(32, 4, activation="relu", padding="same")(x)

    # Layer 9
    x = Flatten()(x)
    x = Dense(512, activation="relu")(x)

    # Output
    x = Dense(n_classes, activation="softmax")(x)
    return Model(inputs=inputs, outputs=x)



from abc import ABC, abstractmethod
from typing import Dict

import pytorch_lightning as pl
import torch

from ertk.pytorch.utils import MTLTaskConfig


class MTLModel(pl.LightningModule, ABC):
    tasks: Dict[str, MTLTaskConfig]

    def __init__(self, tasks: Dict[str, MTLTaskConfig], **kwargs) -> None:
        super().__init__(**kwargs)

        self.tasks = tasks

    @abstractmethod
    def forward(self, x: torch.Tensor, **kwargs) -> Dict[str, torch.Tensor]:
        raise NotImplementedError()

    def training_step(self, batch, batch_idx: int) -> torch.Tensor:
        x, ys = batch
        outputs = self.forward(x)
        loss = 0
        losses = {k: 0 for k in self.tasks}
        for k, v in outputs.items():
            losses[k] += self.tasks[k].loss(v, ys[k]) * self.tasks[k].weight
            loss += losses[k]
        res = dict(loss=loss, **{f"train_{k}_loss": losses[k].detach() for k in losses})
        self.log_dict(res)
        return loss

    def validation_step(self, batch, batch_idx: int) -> None:
        x, ys = batch
        outputs = self.forward(x)
        val_loss = 0
        accs = {k: 0 for k in self.tasks}
        losses = {k: 0 for k in self.tasks}
        for k, v in outputs.items():
            losses[k] += self.tasks[k].loss(v, ys[k]).item() * self.tasks[k].weight
            val_loss += losses[k]
            y_pred = v.argmax(1)
            accs[k] = torch.sum(y_pred == ys[k]).item() / float(len(ys[k]))
        self.log_dict(
            dict(
                val_loss=val_loss,
                **{f"{k}_acc": accs[k] for k in accs},
                **{f"{k}_loss": losses[k] for k in losses},
            )
        )



"""Implementation of the model from [1].

[1] Z. Zhang, B. Wu, and B. Schuller, 'Attention-augmented End-to-end
Multi-task Learning for Emotion Prediction from Speech', in ICASSP 2019
- 2019 IEEE International Conference on Acoustics, Speech and Signal
Processing (ICASSP), May 2019, pp. 6705–6709, doi:
10.1109/ICASSP.2019.8682896.
"""

from typing import Optional

import numpy as np
import tensorflow as tf
from tensorflow.keras.layers import (
    GRU,
    RNN,
    Conv1D,
    Dense,
    Dropout,
    GRUCell,
    Input,
    MaxPool1D,
    Reshape,
    TimeDistributed,
)
from tensorflow.keras.models import Model, Sequential

from ertk.tensorflow.utils import tf_dataset_mem

from .layers import Attention1D

__all__ = ["model", "create_windowed_dataset"]


def create_windowed_dataset(
    x: np.ndarray,
    y: np.ndarray,
    sample_weight: Optional[np.ndarray] = None,
    batch_size: int = 64,
    shuffle: bool = True,
) -> tf.data.Dataset:
    """Creates a non-ragged dataset with zero-padding."""
    arrs = []
    for seq in x:
        seq = np.squeeze(seq)
        arr = np.zeros((500, 640), dtype=np.float32)
        for i in range(0, min(len(seq), 80000), 160):
            idx = i // 160
            maxl = min(len(seq) - i, 640)
            arr[idx, :maxl] = seq[i : i + 640]
        arrs.append(arr)
    arrs = np.array(arrs)
    assert tuple(arrs.shape) == (len(x), 500, 640)
    return tf_dataset_mem(
        arrs, y, sample_weight=sample_weight, batch_size=batch_size, shuffle=shuffle
    )


def model(n_classes: int, n_features: int = 1):
    # Input dimensionality is 640 'features' which are actually 640
    # samples per 40ms segment. Each subsequent vector is shifted 10ms
    # from the previous. We assume the sequences are zero-padded to a
    # multiple of 640 samples.
    inputs = Input((500, 640), name="input")
    frames = tf.expand_dims(inputs, -1)
    dropout = Dropout(0.1)(frames)

    # Conv + maxpool over the 640 samples
    extraction = Sequential()
    extraction.add(Conv1D(40, 40, padding="same", activation="relu"))
    extraction.add(MaxPool1D(2))
    extraction.add(Conv1D(40, 40, padding="same", activation="relu"))
    extraction.add(MaxPool1D(10, data_format="channels_first"))
    extraction.add(Reshape((1280,)))
    features = TimeDistributed(extraction)(dropout)

    # Sequence modelling
    # gru = GRU(128, return_sequences=True)(features)
    gru = RNN([GRUCell(128), GRUCell(128)], return_sequences=True)(features)

    # Attention
    att = Attention1D()(gru)

    x = Dense(n_classes, activation="softmax")(att)
    return Model(inputs=inputs, outputs=x)



import logging
import warnings
from collections import defaultdict
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import pytorch_lightning as pl
import torch
import torch.nn as nn
from pytorch_lightning.loggers import CSVLogger, TensorBoardLogger
from sklearn.model_selection import BaseCrossValidator, LeaveOneGroupOut
from torch.utils.data import DataLoader, TensorDataset

from ertk.pytorch.dataset import MTLDataModule
from ertk.pytorch.models.mtl import MTLModel
from ertk.pytorch.utils import LightningWrapper
from ertk.train import get_scores
from ertk.utils import ScoreFunction


def pt_cross_validate(
    model_fn: Callable[..., Union[nn.Module, pl.LightningModule]],
    x: np.ndarray,
    y: np.ndarray,
    *,
    groups: Optional[np.ndarray] = None,
    cv: BaseCrossValidator = LeaveOneGroupOut(),
    scoring: Union[
        str, List[str], Dict[str, ScoreFunction], Callable[..., float]
    ] = "accuracy",
    verbose: int = 0,
    fit_params: Dict[str, Any] = {},
):
    batch_size = fit_params.pop("batch_size")
    sw = fit_params.pop("sample_weight", None)

    logging.debug(f"cv={cv}")
    logging.debug(f"sample_weight={sw}")
    logging.debug(f"fit_params={fit_params}")

    n_folds = cv.get_n_splits(x, y, groups)
    scores = defaultdict(list)
    for fold, (train, test) in enumerate(cv.split(x, y, groups)):
        if verbose:
            logging.info(f"Fold {fold + 1}/{n_folds}")

        x_train = x[train]
        y_train = y[train]
        x_test = x[test]
        y_test = y[test]
        sw_train = sw[train] if sw is not None else None
        sw_test = sw[test] if sw is not None else None

        _scores = pt_train_val_test(
            model_fn,
            train_data=(x_train, y_train, sw_train),
            valid_data=(x_test, y_test, sw_test),
            test_data=(x_test, y_test, sw_test),
            scoring=scoring,
            verbose=verbose,
            fit_params=dict(fit_params),
            batch_size=batch_size,
        )

        for k in _scores:
            scores[k].append(_scores[k])
    return {k: np.array(scores[k]) for k in scores}


def pt_train_val_test(
    model_fn: Callable[..., Union[nn.Module, pl.LightningModule]],
    train_data: Tuple[np.ndarray, ...],
    valid_data: Tuple[np.ndarray, ...],
    test_data: Optional[Tuple[np.ndarray, ...]] = None,
    scoring: Union[
        str, List[str], Dict[str, ScoreFunction], Callable[..., float]
    ] = "accuracy",
    verbose: int = 0,
    batch_size: int = 32,
    fit_params: Dict[str, Any] = None,
):
    fit_params = fit_params or {}
    log_dir = fit_params.pop("log_dir", None)
    batch_size = fit_params.pop("batch_size", batch_size)
    n_gpus = fit_params.pop("n_gpus", 1)

    logging.debug(f"log_dir={log_dir}")

    logger: Union[bool, List] = False
    if log_dir is not None:
        logger = [
            TensorBoardLogger(log_dir, name=None, version=""),
            CSVLogger(log_dir, name=None, version=""),
        ]

    if test_data is None:
        test_data = valid_data

    clf = model_fn()
    logging.debug(clf)
    if isinstance(clf, tuple):
        transform = clf[0]
        clf = clf[1]
        transform.fit(train_data[0], y=train_data[1])
        train_data = (transform.transform(train_data[0]), *train_data[1:])
        valid_data = (transform.transform(valid_data[0]), *valid_data[1:])
        test_data = (transform.transform(test_data[0]), *test_data[1:])

    if not isinstance(clf, pl.LightningModule):
        clf = LightningWrapper(
            clf,
            nn.CrossEntropyLoss(reduction="none"),
            optim_fn=fit_params.pop("optim_fn"),
        )

    train_data_pt = tuple(torch.from_numpy(x) for x in train_data)
    valid_data_pt = tuple(torch.from_numpy(x) for x in valid_data)
    test_data_pt = tuple(torch.from_numpy(x) for x in test_data)

    train_ds = TensorDataset(*train_data_pt)
    train_dl = DataLoader(
        train_ds, batch_size=batch_size, shuffle=True, pin_memory=True
    )
    valid_ds = TensorDataset(*valid_data_pt)
    valid_dl = DataLoader(
        valid_ds, batch_size=batch_size, shuffle=False, pin_memory=True
    )
    test_ds = TensorDataset(*test_data_pt)
    test_dl = DataLoader(test_ds, batch_size=batch_size, shuffle=False, pin_memory=True)

    if verbose <= 0:
        logging.getLogger("pytorch_lightning").setLevel(logging.ERROR)
        warnings.simplefilter("ignore", UserWarning)
    trainer = pl.Trainer(
        gpus=n_gpus,
        strategy="dp",
        max_epochs=fit_params.pop("max_epochs"),
        logger=logger,
        enable_progress_bar=bool(verbose),
        enable_model_summary=bool(verbose),
        enable_checkpointing=False,
    )
    trainer.fit(clf, train_dataloaders=train_dl, val_dataloaders=valid_dl)
    y_pred = torch.argmax(torch.cat(trainer.predict(clf, test_dl)), -1).cpu().numpy()
    warnings.simplefilter("default", UserWarning)

    y_true = test_data[1]
    scores = get_scores(scoring, y_pred, y_true)
    scores = {f"test_{k}": v for k, v in scores.items()}
    return scores


def train_mtl_model(
    model: MTLModel,
    tasks: Dict[int, int],
    data: Optional[MTLDataModule],
    train_data: Optional[DataLoader] = None,
    valid_data: Optional[DataLoader] = None,
    epochs: int = 200,
    device: str = "cuda:0",
    verbose: int = 0,
):
    trainer = pl.Trainer(
        gpus=1,
        max_epochs=epochs,
        logger=bool(verbose),
        enable_progress_bar=bool(verbose),
        enable_model_summary=bool(verbose),
        enable_checkpointing=False,
    )
    cols = [f"{x}_acc" for x in tasks] + [f"{x}_loss" for x in tasks]
    stats_df = pd.DataFrame(
        columns=pd.Index(cols, name="metric"),
        index=pd.MultiIndex.from_product(
            [range(epochs), ["train", "val"]], names=["epoch", "phase"]
        ),
    )

    if data is not None:
        trainer.fit(model, data)
        [val] = trainer.validate(model, data, verbose=verbose)
    else:
        trainer.fit(model, train_dataloaders=train_data, val_dataloaders=valid_data)
        [val] = trainer.validate(model, valid_data, verbose=verbose)

    return val



import pickle
from pathlib import Path

import click
import numpy as np
import pandas as pd

from ertk.dataset import read_features


@click.command()
@click.argument("input", type=click.Path(exists=True, dir_okay=False, path_type=Path))
@click.argument("output", type=Path)
@click.option(
    "--model",
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    required=True,
    help="Pickled model.",
)
@click.option(
    "--norm",
    type=click.Choice(["speaker", "corpus", "all"]),
    default="speaker",
    show_default=True,
    help="Normalisation scheme.",
)
def main(input: Path, output: Path, model: Path, norm: str):
    """Perform inference on given INPUT features using a pre-trained
    model. Gives the predicted class and confidence and writes CSV to
    OUTPUT.
    """

    dataset = read_features(input)
    names = np.array(dataset.names)
    with open(model, "rb") as fid:
        print(f"Loading model from {model}")
        clf = pickle.load(fid)

    print(f"Running inference on {len(dataset)} clips")
    if hasattr(clf, "predict_proba"):
        pred = clf.predict_proba(dataset.features)
        labels = np.argmax(pred, axis=-1)
        scores = pred[np.arange(pred.shape[0]), labels]
    else:
        labels = clf.predict(dataset.features)
        scores = np.zeros(len(names))

    output.parent.mkdir(exist_ok=True, parents=True)
    df = pd.DataFrame({"Clip": names, "Label": labels, "Score": scores})
    df.to_csv(output, header=True, index=False)
    print(f"Wrote CSV to {output}")


if __name__ == "__main__":
    main()



import click

from .annotation_stats import main as annotation_stats_main
from .combine import main as combine_main
from .convert import main as convert_main
from .info import main as info_main
from .process import main as process_main
from .remove_instances import main as remove_instances_main
from .view_features import main as view_features_main


@click.group(no_args_is_help=True)
def cli():
    """ERTK dataset-related scripts."""
    click.echo("ERTK dataset CLI")


cli.add_command(annotation_stats_main, "annotation")
cli.add_command(combine_main, "combine")
cli.add_command(convert_main, "convert")
cli.add_command(info_main, "info")
cli.add_command(process_main, "process")
cli.add_command(remove_instances_main, "remove_instances")
cli.add_command(view_features_main, "view")



from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, List

import numpy as np
import opensmile
from omegaconf import MISSING

from ertk.config import ERTKConfig

from .base import AudioClipProcessor, FeatureExtractor


@dataclass
class OpenSMILEExtractorConfig(ERTKConfig):
    opensmile_config: str = MISSING
    levels: List[str] = field(default_factory=lambda: ["func"])
    opensmile_opts: Dict[str, str] = field(default_factory=dict)


class OpenSMILEExtractor(
    FeatureExtractor,
    AudioClipProcessor,
    fname="opensmile",
    config=OpenSMILEExtractorConfig,
):
    config: OpenSMILEExtractorConfig

    def __init__(self, config: OpenSMILEExtractorConfig) -> None:
        super().__init__(config)

        opensmile_config = config.opensmile_config
        if not Path(opensmile_config).exists():
            opensmile_config = opensmile.FeatureSet[opensmile_config]

        self.levels = config.levels
        self.smiles = [
            opensmile.Smile(
                opensmile_config, level, options=dict(config.opensmile_opts)
            )
            for level in config.levels
        ]
        self.feature_names = []
        for smile in self.smiles:
            self.feature_names.extend(smile.feature_names)

    def process_instance(self, x: np.ndarray, **kwargs) -> np.ndarray:
        xs = [smile(x, kwargs.pop("sr")) for smile in self.smiles]
        # TODO: allow other window/padding correction options
        cut = min(x.shape[2] for x in xs)
        xs = [x[:, :, :cut] for x in xs]
        x = np.concatenate(xs, axis=1)
        # Get single segment and transpose axes so that frames are first
        return x[0].T

    @property
    def dim(self) -> int:
        return len(self.feature_names)

    @property
    def is_sequence(self) -> bool:
        return "func" not in self.levels



from pathlib import Path

import click
import yaml
from sklearn.model_selection import ParameterGrid

from ertk.config import get_arg_mapping


@click.command()
@click.argument(
    "param_grid", type=click.Path(exists=True, dir_okay=False, path_type=Path)
)
@click.argument("output", type=Path)
@click.option("--format", help="Format string.")
def main(param_grid: Path, output: Path, format: str):
    """Creates a new parameters YAML file in the OUTPUT directory for
    each combination of parameters in the PARAM_GRID file. The names of
    the files will be formatted according to the --format parameter if
    given, or else assigned a number starting from 1.
    """
    grid = get_arg_mapping(param_grid)
    output.mkdir(exist_ok=True, parents=True)
    for i, params in enumerate(ParameterGrid(grid)):
        if format:
            filename = format.format(**params)
        else:
            filename = f"params_{i:02d}"
        if not filename.endswith(".yaml"):
            filename += ".yaml"
        with open(output / filename, "w") as fid:
            yaml.dump(params, fid)
            print(f"Wrote {output / filename}.")


if __name__ == "__main__":
    main()



import shutil
from collections import defaultdict
from pathlib import Path
from typing import Dict, List

import click

from ertk.dataset import Dataset, get_audio_paths


@click.command()
@click.argument("input", type=click.Path(exists=True, dir_okay=False, path_type=Path))
@click.argument("output", type=Path)
@click.option(
    "--label", "label_name", default="label", help="Categorical label to use."
)
def main(input: Path, output: Path, label_name: str):
    """Create directory structure with speaker-independent
    cross-validation folds. Each speaker has a directory which is
    patitioned by label.
    """

    dataset = Dataset(input)
    paths = get_audio_paths(dataset._subset_paths[dataset.subset])
    speaker_paths: Dict[str, List[Path]] = defaultdict(list)
    for path in paths:
        speaker_paths[dataset.annotations["speaker"][path.stem]].append(path)

    for i, speaker in enumerate(speaker_paths.keys()):
        for path in speaker_paths[speaker]:
            label = dataset.annotations[label_name][path.stem]
            fold = f"fold_{i + 1:d}"
            newpath = output / fold / label / path.name
            newpath.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy(str(path), str(newpath))
            print(newpath)


if __name__ == "__main__":
    main()



from collections import OrderedDict
from typing import Dict, Sequence, Type

import torch
import torch.nn as nn
import torch.nn.functional as F

from .mtl import MTLModel, MTLTaskConfig


def make_hidden(sizes: Sequence[int], activation: Type[nn.Module]) -> nn.Module:
    if len(sizes) == 0:
        return nn.Identity()
    return nn.Sequential(
        *(
            nn.Sequential(
                nn.Linear(sizes[i], sizes[i + 1]),
                activation(),
            )
            for i in range(len(sizes) - 1)
        )
    )


class SimpleMTLModel(MTLModel):
    def __init__(
        self,
        tasks: Dict[str, MTLTaskConfig],
        input_dim: int,
        opt_lr: float = 0.001,
        shared_units: Sequence[int] = [512],
        task_units: Sequence[int] = [512],
    ):
        super().__init__(tasks=tasks)

        self.opt_lr = opt_lr

        self.save_hyperparameters("opt_lr", "shared_units", "task_units")

        shared_sizes = [input_dim] + list(shared_units)
        self.shared = make_hidden(shared_sizes, nn.ReLU)
        self.add_module("shared", self.shared)

        self.task_layer = {}
        self.task_output = {}
        task_sizes = [shared_sizes[-1]] + list(task_units)
        for name, task in self.tasks.items():
            self.task_layer[name] = make_hidden(task_sizes, nn.ReLU)
            self.task_output[name] = nn.Linear(task_sizes[-1], task.output_dim)
            submod = nn.Sequential(
                OrderedDict(
                    {
                        "embeddings": self.task_layer[name],
                        "output": self.task_output[name],
                    }
                )
            )
            self.add_module(name, submod)

    def forward(self, x: torch.Tensor, **kwargs) -> Dict[str, torch.Tensor]:
        x = F.relu(self.shared(x))
        task_embeddings = {k: self.task_layer[k](x) for k in self.tasks}
        if kwargs.pop("embeddings_only", False):
            return task_embeddings
        out = {k: F.relu(v) for k, v in task_embeddings.items()}
        return {k: self.task_output[k](out[k]) for k in self.tasks}

    def configure_optimizers(self):
        return torch.optim.Adam(self.parameters(), lr=self.opt_lr)



# Copyright 2019 The TensorFlow Authors All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Core model definition of YAMNet."""

import csv

import numpy as np
import tensorflow as tf
from tensorflow.keras import Model, layers

import features as features_lib


def _batch_norm(name, params):
    def _bn_layer(layer_input):
        return layers.BatchNormalization(
            name=name,
            center=params.batchnorm_center,
            scale=params.batchnorm_scale,
            epsilon=params.batchnorm_epsilon,
        )(layer_input)

    return _bn_layer


def _conv(name, kernel, stride, filters, params):
    def _conv_layer(layer_input):
        output = layers.Conv2D(
            name="{}/conv".format(name),
            filters=filters,
            kernel_size=kernel,
            strides=stride,
            padding=params.conv_padding,
            use_bias=False,
            activation=None,
        )(layer_input)
        output = _batch_norm("{}/conv/bn".format(name), params)(output)
        output = layers.ReLU(name="{}/relu".format(name))(output)
        return output

    return _conv_layer


def _separable_conv(name, kernel, stride, filters, params):
    def _separable_conv_layer(layer_input):
        output = layers.DepthwiseConv2D(
            name="{}/depthwise_conv".format(name),
            kernel_size=kernel,
            strides=stride,
            depth_multiplier=1,
            padding=params.conv_padding,
            use_bias=False,
            activation=None,
        )(layer_input)
        output = _batch_norm("{}/depthwise_conv/bn".format(name), params)(output)
        output = layers.ReLU(name="{}/depthwise_conv/relu".format(name))(output)
        output = layers.Conv2D(
            name="{}/pointwise_conv".format(name),
            filters=filters,
            kernel_size=(1, 1),
            strides=1,
            padding=params.conv_padding,
            use_bias=False,
            activation=None,
        )(output)
        output = _batch_norm("{}/pointwise_conv/bn".format(name), params)(output)
        output = layers.ReLU(name="{}/pointwise_conv/relu".format(name))(output)
        return output

    return _separable_conv_layer


_YAMNET_LAYER_DEFS = [
    # (layer_function, kernel, stride, num_filters)
    (_conv, [3, 3], 2, 32),
    (_separable_conv, [3, 3], 1, 64),
    (_separable_conv, [3, 3], 2, 128),
    (_separable_conv, [3, 3], 1, 128),
    (_separable_conv, [3, 3], 2, 256),
    (_separable_conv, [3, 3], 1, 256),
    (_separable_conv, [3, 3], 2, 512),
    (_separable_conv, [3, 3], 1, 512),
    (_separable_conv, [3, 3], 1, 512),
    (_separable_conv, [3, 3], 1, 512),
    (_separable_conv, [3, 3], 1, 512),
    (_separable_conv, [3, 3], 1, 512),
    (_separable_conv, [3, 3], 2, 1024),
    (_separable_conv, [3, 3], 1, 1024),
]


def yamnet(features, params):
    """Define the core YAMNet mode in Keras."""
    net = layers.Reshape(
        (params.patch_frames, params.patch_bands, 1),
        input_shape=(params.patch_frames, params.patch_bands),
    )(features)
    for (i, (layer_fun, kernel, stride, filters)) in enumerate(_YAMNET_LAYER_DEFS):
        net = layer_fun("layer{}".format(i + 1), kernel, stride, filters, params)(net)
    embeddings = layers.GlobalAveragePooling2D()(net)
    logits = layers.Dense(units=params.num_classes, use_bias=True)(embeddings)
    predictions = layers.Activation(activation=params.classifier_activation)(logits)
    return predictions, embeddings


def yamnet_frames_model(params):
    """Defines the YAMNet waveform-to-class-scores model.

    Args:
      params: An instance of Params containing hyperparameters.

    Returns:
      A model accepting (num_samples,) waveform input and emitting:
      - predictions: (num_patches, num_classes) matrix of class scores per time frame
      - embeddings: (num_patches, embedding size) matrix of embeddings per time frame
      - log_mel_spectrogram: (num_spectrogram_frames, num_mel_bins) spectrogram feature matrix
    """
    waveform = layers.Input(batch_shape=(None,), dtype=tf.float32)
    waveform_padded = features_lib.pad_waveform(waveform, params)
    (
        log_mel_spectrogram,
        features,
    ) = features_lib.waveform_to_log_mel_spectrogram_patches(waveform_padded, params)
    predictions, embeddings = yamnet(features, params)
    frames_model = Model(
        name="yamnet_frames",
        inputs=waveform,
        outputs=[predictions, embeddings, log_mel_spectrogram],
    )
    return frames_model


def class_names(class_map_csv):
    """Read the class name definition file and return a list of strings."""
    if tf.is_tensor(class_map_csv):
        class_map_csv = class_map_csv.numpy()
    with open(class_map_csv) as csv_file:
        reader = csv.reader(csv_file)
        next(reader)  # Skip header
        return np.array([display_name for (_, _, display_name) in reader])



from functools import partial
from typing import Any, Dict, Union

from sklearn.base import BaseEstimator
from sklearn.metrics.pairwise import kernel_metrics
from sklearn.svm import SVC


def _get_kernel_func(
    kernel: str,
    degree: int = 3,
    gamma: Union[str, float, None] = "auto",
    coef0: float = 0.0,
):
    """Get the kernel function, with parameters, to use in fit() and
    predict(). This is calculated at runtime in order to more easily
    handle changes in parameters such as kernel, gamma, etc.
    """
    f = kernel_metrics()[kernel]
    params = {}
    gamma = None if gamma == "auto" else gamma
    if kernel == "poly":
        params = {"degree": degree, "coef0": coef0, "gamma": gamma}
    elif kernel == "rbf":
        params = {"gamma": gamma}
    return partial(f, **params)


class PrecomputedSVC(SVC):
    """Class that wraps scikit-learn's SVC to precompute the kernel
    values in order to speed up training. The kernel parameter is a
    string which is transparently mapped to and from the corresponding
    callable function with the relevant parameters (degree, gamma,
    coef0). All other parameters are passed directly to SVC.
    """

    def __init__(
        self,
        *,
        C=1.0,
        kernel="rbf",
        degree=3,
        gamma="auto",
        coef0=0.0,
        shrinking=True,
        probability=False,
        tol=1e-3,
        cache_size=200,
        class_weight=None,
        verbose=False,
        max_iter=-1,
        decision_function_shape="ovr",
        break_ties=False,
        random_state=None,
    ):
        super().__init__(
            C=C,
            kernel=kernel,
            degree=degree,
            gamma=gamma,
            coef0=coef0,
            shrinking=shrinking,
            probability=probability,
            tol=tol,
            cache_size=cache_size,
            class_weight=class_weight,
            verbose=verbose,
            max_iter=max_iter,
            decision_function_shape=decision_function_shape,
            break_ties=break_ties,
            random_state=random_state,
        )
        self.kernel_name = kernel
        self.kernel = _get_kernel_func(
            self.kernel_name, degree=self.degree, gamma=self.gamma, coef0=self.coef0
        )

    def get_params(self, deep: bool = True) -> Dict[str, Any]:
        params = super().get_params(deep=deep)
        params["kernel"] = self.kernel_name
        return params

    def set_params(self, **params) -> BaseEstimator:
        super().set_params(**params)
        if "kernel" in params:
            self.kernel_name = params["kernel"]
        self.kernel = _get_kernel_func(
            self.kernel_name, degree=self.degree, gamma=self.gamma, coef0=self.coef0
        )
        return self



import json
import warnings
from functools import partial
from pathlib import Path
from typing import Tuple

import click
import numpy as np
import pandas as pd
from click_option_group import optgroup
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import MinMaxScaler, StandardScaler

from ertk.classification import dataset_cross_validation, get_balanced_sample_weights
from ertk.config import get_arg_mapping
from ertk.dataset import load_multiple
from ertk.train import get_cv_splitter
from ertk.transform import SequenceTransformWrapper


@click.command()
@click.argument(
    "input", type=click.Path(exists=True, dir_okay=False, path_type=Path), nargs=-1
)
@click.option("--features", required=True, help="Features to load.")
@click.option("--clf", "clf_type", required=True, help="Classifier to use.")
@optgroup.group("Results options")
@optgroup.option("--results", type=Path, help="Results directory.")
@optgroup.option("--logdir", type=Path, help="TF/PyTorch logs directory.")
@optgroup.group("Data options")
@optgroup.option("--label", default="label", help="Label annotation to use.")
@optgroup.option(
    "--subset", multiple=True, default=["default"], help="Subset selection."
)
@optgroup.option("--map_groups", multiple=True, help="Group name mapping.")
@optgroup.option(
    "--sel_groups",
    multiple=True,
    help="Group selection. This is a map from partition to group(s).",
)
@optgroup.option(
    "--clip_seq", type=int, help="Clip sequences to this length (before pad)."
)
@optgroup.option(
    "--pad_seq", type=int, help="Pad sequences to multiple of this length (after clip)."
)
@optgroup.group("Cross-validation options")
@optgroup.option("--partition", help="Partition for LOGO CV.")
@optgroup.option(
    "--kfold",
    type=int,
    default=5,
    help="k when using (group) k-fold cross-validation, or leave-one-out.",
)
@optgroup.option(
    "--inner_kfold",
    type=int,
    default=2,
    help="k for inner k-fold CV (where relevant). If -1 then LOGO is used. If 1 then a "
    "random split is used.",
)
@optgroup.option("--test_size", type=float, help="Test size when kfold=1.")
@optgroup.option(
    "--inner_part", help="Which partition to use for group-based inner CV."
)
@optgroup.group("Training options")
@optgroup.option("--learning_rate", type=float, default=1e-4, show_default=True)
@optgroup.option("--batch_size", type=int, default=64, show_default=True)
@optgroup.option("--epochs", type=int, default=50, show_default=True)
@optgroup.option(
    "--balanced/--imbalanced", default=True, help="Balances sample weights."
)
@optgroup.option(
    "--n_gpus", type=int, default=1, show_default=True, help="Number of GPUs to use."
)
@optgroup.option(
    "--reps",
    type=int,
    default=1,
    show_default=True,
    help="The number of repetitions to do per test.",
)
@optgroup.option(
    "--normalise",
    default="online",
    show_default=True,
    help="Normalisation method. 'online' means use training data for normalisation.",
)
@optgroup.option(
    "--transform",
    type=click.Choice(["std", "minmax"]),
    default="std",
    show_default=True,
    help="Transformation class.",
)
@optgroup.option(
    "--n_jobs", type=int, default=-1, help="Number of parallel executions."
)
@optgroup.group("Misc. options")
@optgroup.option(
    "--verbose",
    type=int,
    default=0,
    help="Verbosity. -1=nothing, 0=dataset+results, 1=INFO, 2=DEBUG",
)
@optgroup.group("Model-specific options")
@optgroup.option(
    "--clf_args",
    "clf_args_file",
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    multiple=True,
    help="File containing keyword arguments to give to model initialisation.",
)
@optgroup.option(
    "--param_grid",
    "param_grid_file",
    type=click.Path(exists=True, path_type=Path),
    multiple=True,
    help="File with parameter grid data.",
)
@optgroup.group(
    "Deprecated options", help="These options are deprecated, don't use them."
)
@optgroup.option(
    "--inner_cv/--noinner_cv",
    "use_inner_cv",
    default=True,
    help="Whether to use inner CV. This is deprecated and only exists for backwards "
    "compatibility.",
)
def main(
    clf_type: str,
    input: Tuple[Path],
    features: str,
    partition: str,
    label: str,
    kfold: int,
    inner_kfold: int,
    test_size: float,
    inner_part: str,
    results: Path,
    logdir: Path,
    reps: int,
    normalise: str,
    transform: str,
    balanced: bool,
    subset: Tuple[str],
    map_groups: Tuple[str],
    sel_groups: Tuple[str],
    clip_seq: int,
    pad_seq: int,
    verbose: int,
    clf_args_file: Tuple[Path],
    param_grid_file: Tuple[Path],
    learning_rate: float,
    batch_size: int,
    epochs: int,
    use_inner_cv: bool,
    n_jobs: int,
    n_gpus: int,
):
    """Runs cross-validation on the given INPUT datasets. Metrics are
    optionally written to a results file.
    """

    np.set_printoptions(precision=4, threshold=10)
    if verbose > 0:
        import logging

        logging.basicConfig(level=logging.DEBUG if verbose > 1 else logging.INFO)
        np.set_printoptions()

    if len(subset) == 1:
        dataset = load_multiple(input, features, subsets=subset[0])
    else:
        subset_map = {}
        for m in subset:
            subset_map.update(get_arg_mapping(m))
        dataset = load_multiple(input, features, subsets=subset_map)

    for m in map_groups:
        mapping = get_arg_mapping(m)
        for part in mapping:
            if part not in dataset.partitions:
                warnings.warn(
                    f"Partition {part} cannot be mapped as it is not in the dataset."
                )
                continue
            dataset.map_groups(part, mapping[part])
    for m in sel_groups:
        mapping = get_arg_mapping(m)
        for part in mapping:
            if part not in dataset.partitions:
                warnings.warn(
                    f"Partition {part} cannot be selected as it is not in the dataset."
                )
                continue
            keep = mapping[part]
            if isinstance(keep, str):
                keep = [keep]
            dataset.remove_groups(part, keep=keep)

    if clip_seq:
        dataset.clip_arrays(clip_seq)
    if pad_seq:
        dataset.pad_arrays(pad_seq)

    transformer = {"std": StandardScaler, "minmax": MinMaxScaler}[transform]()
    if normalise == "none":
        transformer = None
    elif normalise == "online" and len(dataset.x[0].shape) > 1:
        transformer = SequenceTransformWrapper(transformer, "feature")
    elif normalise != "online":
        dataset.normalise(normaliser=transformer, partition=normalise)
        transformer = None

    sample_weight = None
    if balanced:
        sample_weight = get_balanced_sample_weights(dataset.get_annotations(label))
    fit_params = {"sample_weight": sample_weight}

    model_args = {}
    for file in clf_args_file:
        model_args.update(get_arg_mapping(file))
    param_grid = {}
    for file in param_grid_file:
        param_grid.update(get_arg_mapping(file))

    cv = get_cv_splitter(bool(partition), kfold, test_size=test_size)
    if not use_inner_cv:
        warnings.warn(
            "--noinner_cv is deprecated and only exists for backwards compatibility."
        )
        inner_cv = cv
    else:
        inner_cv = get_cv_splitter(bool(inner_part), inner_kfold)

    clf_lib, clf_type = clf_type.split("/")
    if clf_lib == "sk":
        from sklearn.model_selection import GridSearchCV

        from ertk.sklearn.models import get_sk_model

        clf = get_sk_model(clf_type, **model_args)
        clf = GridSearchCV(
            Pipeline([("transform", transformer), ("clf", clf)]),
            {f"clf__{k}": v for k, v in param_grid.items()},
            scoring="balanced_accuracy",
            cv=inner_cv,
            verbose=max(verbose, 0),
            n_jobs=1 if use_inner_cv else n_jobs,
        )
        if inner_part:
            fit_params["groups"] = dataset.get_group_indices(inner_part)
        params = param_grid
        if not use_inner_cv:
            # Get best params then pass best to main cross-validation routine
            # XXX: This exists because of mistake in our original implementation.
            clf.fit(
                dataset.x,
                dataset.labels,
                groups=dataset.get_group_indices(partition) if partition else None,
                clf__sample_weight=sample_weight,
            )
            params = clf.best_params_
            clf = clf.best_estimator_
    elif clf_lib == "tf":
        from tensorflow.keras.optimizers import Adam

        from ertk.tensorflow import get_tf_model
        from ertk.tensorflow.classification import tf_classification_metrics

        # No parallel CV to avoid running out of GPU memory
        n_jobs = 1
        fit_params.update(
            {
                "log_dir": logdir,
                "epochs": epochs,
                "batch_size": batch_size,
                "data_fn": None,
                "n_gpus": n_gpus,
            }
        )
        model_args.update(
            {"n_features": dataset.n_features, "n_classes": dataset.n_classes}
        )

        def model_fn():
            model = get_tf_model(clf_type, **model_args)
            model.compile(
                Adam(learning_rate=learning_rate),
                loss="sparse_categorical_crossentropy",
                metrics=tf_classification_metrics(),
            )
            return Pipeline([("transform", transformer), ("clf", model)])

        params = {
            "optimiser": "adam",
            "learning_rate": learning_rate,
            **model_args,
            **fit_params,
        }
        clf = model_fn
    elif clf_lib == "pt":
        import torch.optim

        from ertk.pytorch import get_pt_model

        n_jobs = 1
        optim_fn = partial(torch.optim.Adam, lr=learning_rate)
        fit_params.update(
            {
                "log_dir": logdir,
                "max_epochs": epochs,
                "batch_size": batch_size,
                "optim_fn": optim_fn,
                "n_gpus": n_gpus,
            }
        )
        model_args.update(
            {"n_features": dataset.n_features, "n_classes": dataset.n_classes}
        )

        def model_fn():
            model = get_pt_model(clf_type, **model_args)
            return model if transformer is None else (transformer, model)

        params = {
            "optimiser": "adam",
            "learning_rate": learning_rate,
            **model_args,
            **fit_params,
        }
        clf = model_fn
    else:
        raise ValueError(f"Invalid classifier type {clf_type}")

    if verbose > -1:
        print("== Dataset ==")
        print(dataset)
        print("== Cross-validation settings ==")
        print(f"Using {dataset.n_classes}-class classification.")
        print(f"Using {n_jobs} parallel jobs.")
        print(f"Using classifier {clf}.")
        print(f"Using fit_params={fit_params}")
        print(f"Using {kfold} k-fold CV.")
        if partition:
            print(f"Using {partition} as split partition.")
        if use_inner_cv:
            if inner_part:
                print(f"Using {inner_part} as inner split partition.")
            print(f"Using {inner_cv} as inner CV splitter.")
        if balanced:
            print("Using balanced sample weights.")
        if normalise == "online":
            print("Using 'online' normalisation.")
        else:
            print(f"Normalising globally using {normalise} partition.")

    dfs = []
    for rep in range(1, reps + 1):
        if verbose > -1:
            print(f"Rep {rep}/{reps}")
        df = dataset_cross_validation(
            clf,
            dataset,
            clf_lib=clf_lib,
            partition=partition,
            label=label,
            cv=cv,
            verbose=max(verbose, 0),
            n_jobs=n_jobs,
            fit_params=fit_params,
        )
        df["params"] = [json.dumps(params, default=str)] * len(df)
        dfs.append(df)
    df = pd.concat(dfs, keys=list(range(1, reps + 1)), names=["rep"])
    df["clf"] = f"{clf_lib}/{clf_type}"
    df["corpus"] = dataset.corpus
    df["features"] = Path(features).stem
    if results:
        results.parent.mkdir(parents=True, exist_ok=True)
        df.to_csv(results)
        print(f"Wrote CSV to {results}")
    else:
        print(df)
    if verbose > -1:
        print(df.mean(numeric_only=True).to_string())


if __name__ == "__main__":
    main()



import warnings

import numpy as np
from joblib import Parallel, delayed
from sklearn.base import MetaEstimatorMixin, clone
from sklearn.multiclass import OneVsRestClassifier as _SKOvR
from sklearn.multiclass import _ConstantPredictor
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import LabelBinarizer


# Workaround for OneVsRest
# https://github.com/scikit-learn/scikit-learn/issues/10882#issuecomment-376912896
# https://stackoverflow.com/a/49535681/10044861
def _fit_binary(estimator, X, y, classes=None, **kwargs):
    unique_y = np.unique(y)
    if len(unique_y) == 1:
        if classes is not None:
            if y[0] == -1:
                c = 0
            else:
                c = y[0]
            warnings.warn(
                "Label %s is present in all training examples." % str(classes[c])
            )
        estimator = _ConstantPredictor().fit(X, unique_y)
    else:
        estimator = clone(estimator)
        estimator.fit(X, y, **kwargs)
    return estimator


class OneVsRestClassifier(_SKOvR):
    def fit(self, X, y, **kwargs):
        self.label_binarizer_ = LabelBinarizer(sparse_output=True)
        Y = self.label_binarizer_.fit_transform(y)
        Y = Y.tocsc()
        self.classes_ = self.label_binarizer_.classes_
        columns = (col.toarray().ravel() for col in Y.T)
        self.estimators_ = Parallel(
            n_jobs=self.n_jobs, backend="multiprocessing", verbose=50
        )(
            delayed(_fit_binary)(
                self.estimator,
                X,
                column,
                classes=[
                    "not %s" % self.label_binarizer_.classes_[i],
                    self.label_binarizer_.classes_[i],
                ],
                **kwargs,
            )
            for i, column in enumerate(columns)
        )
        return self


def get_base_estimator(clf):
    """Gets the base estimator of a pipeline or meta-estimator, assuming
    there is only one.

    Parameters
    ----------
    clf: estimator
        An estimator.

    Returns
    -------
    estimator
        The base estimator of `clf`.
    """
    if isinstance(clf, Pipeline):
        if clf._final_estimator == "passthrough":
            raise ValueError("Pipeline final estimator is 'passthrough'.")
        return get_base_estimator(clf._final_estimator)
    if isinstance(clf, MetaEstimatorMixin):
        for attr in ["base_estimator", "estimator"]:
            if hasattr(clf, attr):
                return get_base_estimator(getattr(clf, attr))
        raise RuntimeError("Couldn't get base estimator from meta estimator")
    return clf



from enum import Enum
from math import pi
from pathlib import Path
from typing import Optional, Sequence, Tuple

import tensorflow as tf

from ertk.utils import PathOrStr

__all__ = ["BBRBM", "GBRBM"]


def _sample_bernoulli(p):
    return tf.nn.relu(tf.sign(p - tf.random.uniform(tf.shape(p))))


def _sample_std_normal(p):
    return tf.random.normal()


class DecayType(Enum):
    STEP = "STEP"
    COSINE = "COSINE"
    EXP = "EXP"

    def __str__(self):
        return self.name


class RBM:
    """Restricted Boltzmann Machine

    This implementation is designed around the work of Hinton et. al.
    [1], [2].

    References
    ----------
    [1] G. E. Hinton and R. R. Salakhutdinov, 'Reducing the
    Dimensionality of Data with Neural Networks', Science, vol. 313, no.
    5786, pp. 504-507, Jul. 2006, doi: 10.1126/science.1127647.

    [2] G. E. Hinton, 'A Practical Guide to Training Restricted
    Boltzmann Machines', in Neural Networks: Tricks of the Trade: Second
    Edition, G. Montavon, G. B. Orr, and K.-R. Müller, Eds. Berlin,
    Heidelberg: Springer Berlin Heidelberg, 2012, pp. 599-619.
    """

    def __init__(
        self,
        n_hidden: int,
        input_shape: Tuple[int, ...],
        logdir: PathOrStr = "logs/rbm",
        output_histograms: bool = False,
    ):
        """
        Parameters
        ----------
        n_hidden: int, default = 128
            Number of hidden units
        input_shape: tuple
            Shape of input tensors, either 1D or 2D
        logdir: str or Path, optional
            Path to store TensorBoard logs. Default is 'logs/rbm'.
        output_histograms: bool, optional
            Whether to output histogram information. Can slow down training
            somewhat. Default is False.
        """
        if any(x is None or x < 0 for x in input_shape):
            raise ValueError("input_shape must be fully specified.")

        self.input_shape = tf.TensorShape(input_shape)
        n_visible = tf.reduce_prod(self.input_shape)

        self.n_hidden = n_hidden

        self.h_bias = tf.Variable(tf.zeros((n_hidden,)))
        self.v_bias = tf.Variable(tf.zeros((n_visible,)))
        self.W = tf.Variable(tf.random.normal((n_visible, n_hidden), 0.0, 0.01))

        self.h_bias_velocity = tf.Variable(tf.zeros_like(self.h_bias), trainable=False)
        self.v_bias_velocity = tf.Variable(tf.zeros_like(self.v_bias), trainable=False)
        self.W_velocity = tf.Variable(tf.zeros_like(self.W), trainable=False)

        self.learning_rate = tf.Variable(0.01, trainable=False)
        self.momentum = tf.Variable(0.5, trainable=False)
        self.l2_weight = tf.Variable(0.001, trainable=False)

        self.output_histograms = output_histograms

        self.logging = False
        if logdir:
            self.logging = True
            logdir = Path(logdir)
            self.train_summary_writer = tf.summary.create_file_writer(
                str(logdir / "train")
            )
            self.valid_summary_writer = tf.summary.create_file_writer(
                str(logdir / "valid")
            )

    def free_energy(self, batch: tf.Tensor):
        raise NotImplementedError("free_energy() not implemented.")

    def hidden_prob(self, batch: tf.Tensor):
        raise NotImplementedError("hidden_prob() not implemented.")

    def visible_prob(self, h: tf.Tensor):
        raise NotImplementedError("visible_prob() not implemented.")

    def sample_h(self, p: tf.Tensor):
        raise NotImplementedError("sample_h() not implemented.")

    def sample_v(self, p: tf.Tensor):
        raise NotImplementedError("sample_v() not implemented.")

    def calculate_deltas(self, batch: tf.Tensor):
        raise NotImplementedError("calculate_deltas() not implemented.")

    def forward(self, batch: tf.Tensor):
        """Samples hidden units from input batch.

        Parameters
        ----------
        batch: tensor of shape (batch, n_visible)
            Batch of tensors to process.

        Returns
        -------
        Sample hidden units.
        """
        p_h = self.hidden_prob(batch)
        h = self.sample_h(p_h)
        return h

    def backward(self, h: tf.Tensor):
        p_v = self.visible_prob(h)
        return p_v

    @tf.function
    def train_batch(self, batch: tf.Tensor):
        """Runs one step of Gibbs sampling and gradient ascent with the given
        batch.

        Parameters
        ----------
        batch: tensor of shape (batch, n_visible)
            Batch of tensors to train on.

        Returns
        -------
        Mean square error (MSE) for this batch.
        """
        p_h = self.hidden_prob(batch)
        h = self.sample_h(p_h)
        p_v = self.visible_prob(h)
        p_hd = tf.sigmoid(self.h_bias + tf.matmul(p_v, self.W))

        # v*h.T - v'*h'.T
        pos = tf.matmul(tf.expand_dims(batch, -1), tf.expand_dims(p_h, -2))
        neg = tf.matmul(tf.expand_dims(p_v, -1), tf.expand_dims(p_hd, -2))
        outers = pos - neg
        W_delta = tf.reduce_mean(outers, 0) - self.l2_weight * 2.0 * self.W
        # h - h'
        h_bias_delta = tf.reduce_mean(p_h - p_hd, 0)
        # v - v'
        v_bias_delta = tf.reduce_mean(batch - p_v, 0)

        self.W_velocity.assign(
            self.momentum * self.W_velocity + self.learning_rate * W_delta
        )
        self.h_bias_velocity.assign(
            self.momentum * self.h_bias_velocity + self.learning_rate * h_bias_delta
        )
        self.v_bias_velocity.assign(
            self.momentum * self.v_bias_velocity + self.learning_rate * v_bias_delta
        )

        self.W.assign_add(self.W_velocity)
        self.h_bias.assign_add(self.h_bias_velocity)
        self.v_bias.assign_add(self.v_bias_velocity)

        mse_batch = tf.reduce_mean((p_v - batch) ** 2)
        return mse_batch

    def train(
        self,
        train_data: tf.data.Dataset,
        valid_data: Optional[tf.data.Dataset] = None,
        init_learning_rate: float = 0.01,
        final_learning_rate: float = 0.001,
        learning_rate_decay: Optional[DecayType] = None,
        learning_rate_decay_param: float = 0.5,
        init_momentum: float = 0.5,
        final_momentum: float = 0.9,
        momentum_decay: Optional[DecayType] = None,
        momentum_decay_param: float = 0.5,
        n_epochs: int = 100,
    ):
        """Train this RBM.

        Parameters
        ----------
        train_data: tf.data.Dataset
            Training data to use.
        valid_data: tf.data.Dataset, optional
            Validation data to use for performance measuring.
        init_learning_rate: float
            The initial learning rate to set. Default is 0.01.
        final_learning_rate: float
            The final learning rate to set. Default is 0.001
        learning_rate_decay: DecayType, optional
            Type of learning rate decay. Default is None, meaning no decay.
        learning_rate_decay_param: float
            Parameter for learning rate decay. Use depends on type. For step
            decay, this is the fraction of epochs after which to change
            learning rate.
        init_momentum: float
            The initial momentum to set. Default is 0.5.
        final_momentum: float
            The final momentum to set. Default is 0.9.
        momentum_decay: DecayType, optional
            Type of momentum decay. Default is None, meaning no decay.
        n_epochs: int
            The number of epochs to train for. A single epoch is one pass over
            the training data. Default is 100.
        """
        self.learning_rate.assign(init_learning_rate)
        self.momentum.assign(init_momentum)
        train_size = sum(1 for _ in train_data)

        if learning_rate_decay == DecayType.COSINE:
            lr_coeffA = 0.5 * (init_learning_rate - final_learning_rate)
            lr_coeffB = 0.5 * (init_learning_rate + final_learning_rate)
        elif learning_rate_decay == DecayType.EXP:
            lr_coeffA = init_learning_rate
            lr_coeffB = tf.math.log(init_learning_rate) - tf.math.log(
                final_learning_rate
            )
        if momentum_decay == DecayType.COSINE:
            mom_coeffA = 0.5 * (init_momentum - final_momentum)
            mom_coeffB = 0.5 * (init_momentum + final_momentum)
        elif momentum_decay == DecayType.EXP:
            mom_coeffA = init_momentum
            mom_coeffB = tf.math.log(init_momentum) - tf.math.log(final_momentum)

        for epoch in range(1, n_epochs + 1):
            tf.print(f"Epoch: {epoch}")
            time_frac = float(epoch) / float(n_epochs)

            if (
                learning_rate_decay == DecayType.STEP
                and time_frac > learning_rate_decay_param
            ):
                self.learning_rate.assign(final_learning_rate)
            elif learning_rate_decay == DecayType.COSINE:
                self.learning_rate.assign(
                    lr_coeffA * tf.cos(pi * time_frac) + lr_coeffB
                )
            elif learning_rate_decay == DecayType.EXP:
                self.learning_rate.assign(lr_coeffA * tf.exp(-lr_coeffB * time_frac))

            if momentum_decay == DecayType.STEP and time_frac > momentum_decay_param:
                self.momentum.assign(final_momentum)
            elif momentum_decay == DecayType.COSINE:
                self.momentum.assign(mom_coeffA * tf.cos(pi * time_frac) + mom_coeffB)
            elif momentum_decay == DecayType.EXP:
                self.momentum.assign(mom_coeffA * tf.exp(-mom_coeffB * time_frac))

            mse_train = tf.TensorArray(tf.float32, size=0, dynamic_size=True)
            i = tf.constant(0, tf.int32)
            for batch in train_data:
                if i % (train_size // 10) == 0:
                    tf.print(".", end="")

                mse_batch = self.train_batch(batch)
                mse_train = mse_train.write(i, mse_batch)
                i += 1

            mse_train = mse_train.stack()
            mse_train = tf.reduce_mean(mse_train)

            with self.train_summary_writer.as_default():
                tf.summary.scalar("mse", mse_train, step=epoch)
                tf.print(f"mse: {mse_train:.3f}, ", end="")

                tf.summary.scalar("learining_rate", self.learning_rate, step=epoch)
                tf.summary.scalar("momentum", self.momentum, step=epoch)
                weights_l2 = tf.reduce_sum(self.W ** 2)
                tf.summary.scalar("weights_l2", weights_l2, step=epoch)

                if self.output_histograms:
                    tf.summary.histogram("h_bias", self.h_bias, step=epoch)
                    tf.summary.histogram("v_bias", self.v_bias, step=epoch)
                    tf.summary.histogram("weights", self.W, step=epoch)

                    tf.summary.histogram(
                        "h_bias_velocity", self.h_bias_velocity, step=epoch
                    )
                    tf.summary.histogram(
                        "v_bias_velocity", self.v_bias_velocity, step=epoch
                    )
                    tf.summary.histogram(
                        "weights_velocity", self.W_velocity, step=epoch
                    )

                batch = next(iter(train_data))
                f_energy = self.free_energy(batch)
                tf.summary.scalar("free_energy", f_energy, step=epoch)

                if 2 <= len(self.input_shape) <= 3:
                    if len(self.input_shape) == 3:
                        new_shape = tf.concat(
                            ([self.n_hidden], self.input_shape), axis=0
                        )
                    elif len(self.input_shape) == 2:
                        new_shape = tf.concat(
                            ([self.n_hidden], self.input_shape, [1]), axis=0
                        )
                    W = tf.transpose(self.W)
                    W = tf.reshape(W, new_shape)
                    W = (W + 2.0) / 4.0
                    tf.summary.image("hidden_vis", W, step=epoch, max_outputs=128)

                    if len(self.input_shape) == 3:
                        new_shape = tf.concat(([-1], self.input_shape), axis=0)
                    elif len(self.input_shape) == 2:
                        new_shape = tf.concat(([-1], self.input_shape, [1]), axis=0)
                    reconst = self.reconstruct_batch(batch)
                    reconst = tf.reshape(reconst, new_shape)
                    image = tf.reshape(batch, new_shape)
                    combined = tf.concat([image, reconst], 2)
                    tf.summary.image(
                        "reconstruction", combined, step=epoch, max_outputs=10
                    )

            if valid_data:
                mse_valid = tf.TensorArray(tf.float32, size=0, dynamic_size=True)
                i = tf.constant(0, tf.int32)
                for batch in valid_data:
                    p_v = self.reconstruct_batch(batch)
                    mse_batch = tf.reduce_mean((p_v - batch) ** 2)
                    mse_valid = mse_valid.write(i, mse_batch)
                    i += 1
                mse_valid = mse_valid.stack()
                mse_valid = tf.reduce_mean(mse_valid)

                with self.valid_summary_writer.as_default():
                    tf.summary.scalar("mse", mse_valid, step=epoch)
                    tf.print(f"val_mse: {mse_valid:.3f}")

                    batch = next(iter(valid_data))

                    if 2 <= len(self.input_shape) <= 3:
                        if len(self.input_shape) == 3:
                            new_shape = tf.concat(([-1], self.input_shape), axis=0)
                        elif len(self.input_shape) == 2:
                            new_shape = tf.concat(([-1], self.input_shape, [1]), axis=0)
                        reconst = self.reconstruct_batch(batch)
                        reconst = tf.reshape(reconst, new_shape)
                        image = tf.reshape(batch, new_shape)
                        combined = tf.concat([image, reconst], 2)
                        tf.summary.image(
                            "reconstruction", combined, step=epoch, max_outputs=10
                        )

                    p_h = self.hidden_prob(batch)
                    p_h = tf.reshape(p_h, (1, -1, self.n_hidden, 1))
                    p_h = tf.transpose(p_h, [0, 2, 1, 3])
                    tf.summary.image("hidden_prob", p_h, step=epoch)

                    f_energy = self.free_energy(batch)
                    tf.summary.scalar("free_energy", f_energy, step=epoch)

    @tf.function
    def reconstruct_batch(self, batch):
        """Reconstruct a single batch of input.

        Parameters
        ----------
        batch: tensor of shape (batch, n_visible)
            The batch of tensor to reconstruct.

        Returns
        -------
        Reconstructed batch.
        """
        p_h = self.hidden_prob(batch)
        h = self.sample_h(p_h)
        p_v = self.visible_prob(h)
        return p_v

    def reconstruct(self, data: tf.data.Dataset):
        """Reconstruct all input data.

        Parameters
        ----------
        data: tf.data.Dataset
            Data to reconstruct.

        Returns
        -------
        The reconstructed data as a mapped tf.data.Dataset.
        """
        return data.map(self.reconstruct_batch)

    def representations(self, data: tf.data.Dataset):
        return data.map(self.hidden_prob)


class BBRBM(RBM):
    """RBM with Bernoulli visible and hidden units."""

    def free_energy(self, batch: tf.Tensor):
        x = self.h_bias + tf.matmul(batch, self.W)
        f_energy = -tf.reduce_sum(batch * self.v_bias, -1) - tf.reduce_sum(
            tf.nn.softplus(x), -1
        )
        f_energy = tf.reduce_mean(f_energy)
        return f_energy

    def hidden_prob(self, batch: tf.Tensor):
        return tf.sigmoid(self.h_bias + tf.matmul(batch, self.W))

    def visible_prob(self, h: tf.Tensor):
        return tf.sigmoid(self.v_bias + tf.matmul(h, self.W, transpose_b=True))

    def sample_h(self, p: tf.Tensor):
        return _sample_bernoulli(p)


class GBRBM(RBM):
    """RBM with Gaussian visible and Bernoulli hidden units."""

    def free_energy(self, batch: tf.Tensor):
        x = self.h_bias + tf.matmul(batch, self.W)
        f_energy = 0.5 * tf.reduce_sum((batch - self.v_bias) ** 2, -1) - tf.reduce_sum(
            tf.nn.softplus(x), -1
        )
        f_energy = tf.reduce_mean(f_energy)
        return f_energy

    def hidden_prob(self, batch: tf.Tensor):
        return tf.sigmoid(self.h_bias + tf.matmul(batch, self.W))

    def visible_prob(self, h: tf.Tensor):
        return self.v_bias + tf.matmul(h, self.W, transpose_b=True)

    def sample_h(self, p: tf.Tensor):
        return _sample_bernoulli(p)


class DBN:
    """Deep Belief Network, also known as a Deep Boltzmann Machine."""

    def __init__(
        self,
        input_shape: Tuple[int],
        n_layers: int = 3,
        layer_nodes: Sequence[int] = [512, 256, 128],
        logdir: PathOrStr = "logs/dbn",
        **kwargs,
    ):
        """Initialise this DBN.

        Parameters
        ----------
        input_shape: tuple
            Shape of the input tensors to the first layer
        n_layers: int
            Number of hidden layers in this DBN.
        layer_nodes: tuple of int
            Number of hidden nodes in each of the layers. We must have that
            len(layer_nodes) == n_layers
        logdir: str or Path
            Directory to save log files.
        kwargs: dict
            Additional arguments to pass to the RBM constructor.
        """
        if n_layers != len(layer_nodes):
            raise ValueError("layer_nodes must have exactly n_layers elements")

        self.input_shape = input_shape

        logdir = Path(logdir)
        timestamp = tf.timestamp().numpy()
        logdir = logdir / str(timestamp)
        self.layers = [
            BBRBM(layer_nodes[0], self.input_shape, logdir=logdir / "rbm0", **kwargs)
        ]
        for i in range(1, n_layers):
            self.layers.append(
                BBRBM(
                    layer_nodes[i],
                    logdir=logdir / f"rbm{i}",
                    input_shape=(layer_nodes[i - 1],),
                )
            )

        self.train_summary_writer = tf.summary.create_file_writer(
            str(logdir / "dbn" / "train")
        )
        self.valid_summary_writer = tf.summary.create_file_writer(
            str(logdir / "dbn" / "valid")
        )

    def train(
        self,
        train_data: tf.data.Dataset,
        valid_data: Optional[tf.data.Dataset] = None,
        n_epochs: int = 100,
        learning_rate: float = 0.01,
        momentum: float = 0.5,
    ):
        """Train each layer in this DBN in sequence, using the reconstructed
        output from previous layers as input.

        Parameters
        ----------
        train_data: tf.data.Dataset
            Training data to use.
        valid_data: tf.data.Dataset, optional
            Validation data to use for mesuring training performance.
        n_epochs: int
            The number of epochs to train each layer for. Default is 100.
        learning_rate: float
            The learning rate to use for all layers. Default is 0.01.
        momentum: float
            The momentum to use for all layers. Default is 0.5.
        """
        print("Training layer 0")
        self.layers[0].train(
            train_data,
            valid_data=valid_data,
            init_momentum=momentum,
            init_learning_rate=learning_rate,
            n_epochs=n_epochs,
        )
        new_train_data = train_data
        new_valid_data = valid_data
        for i in range(1, len(self.layers)):
            print(f"Training layer {i}")
            new_train_data = new_train_data.map(self.layers[i - 1].forward)
            if new_valid_data:
                new_valid_data = new_valid_data.map(self.layers[i - 1].forward)
            self.layers[i].train(
                new_train_data,
                valid_data=new_valid_data,
                init_momentum=momentum,
                init_learning_rate=learning_rate,
                n_epochs=n_epochs,
            )

        if valid_data:
            with self.valid_summary_writer.as_default():
                batch = next(iter(valid_data))

                if len(self.input_shape) == 3:
                    new_shape = tf.concat(([-1], self.input_shape), axis=0)
                elif len(self.input_shape) == 2:
                    new_shape = tf.concat(([-1], self.input_shape, [1]), axis=0)
                tf.print("Reconstructing")
                reconst = self.reconstruct_batch(batch)
                reconst = tf.reshape(reconst, new_shape)
                image = tf.reshape(batch, new_shape)
                combined = tf.concat([image, reconst], 2)
                tf.summary.image(
                    "reconstruction", combined, step=n_epochs, max_outputs=10
                )

    @tf.function
    def reconstruct_batch(self, batch: tf.Tensor):
        """Reconstruct a single batch of input.

        Parameters
        ----------
        batch: tensor of shape (batch, n_visible)
            The batch of tensor to reconstruct.

        Returns
        -------
        Reconstructed batch.
        """
        for i in range(len(self.layers)):
            batch = self.layers[i].forward(batch)
        for i in range(len(self.layers) - 1, -1, -1):
            batch = self.layers[i].backward(batch)
        return batch

    def reconstruct(self, data: tf.data.Dataset):
        """Reconstruct all input data.

        Parameters
        ----------
        data: tf.data.Dataset
            Data to reconstruct.

        Returns
        -------
        The reconstructed data as a mapped tf.data.Dataset.
        """
        return data.map(self.reconstruct_batch)



from dataclasses import dataclass
from enum import Enum

import librosa
import numpy as np

from ertk.config import ERTKConfig
from ertk.preprocessing.base import AudioClipProcessor


class _VADMethod(Enum):
    mh2009 = "MH2009"
    librosa = "librosa"


@dataclass
class VADTrimmerConfig(ERTKConfig):
    method: _VADMethod = _VADMethod.librosa
    top_db: int = 60
    energy_thresh: float = 2.5
    freq_thresh: float = 150
    sf_thresh: float = 10
    window: float = 0.01
    debug: bool = False
    min_speech: int = 5
    min_silence: int = 10


def filter_silence_speech(x: np.ndarray, min_speech: int = 5, min_silence: int = 10):
    """Smooths a boolean array indicating segments of speech based on
    the given speech and silence thresholds.
    """
    silence = True
    i = 0
    while i < len(x):
        if silence and x[i]:
            # Going from silence to speech
            j = next((j for j in range(i + 1, len(x)) if not x[j]), len(x))
            if j - i >= min_speech:
                silence = False
            else:
                x[i:j] = False
        elif not silence and not x[i]:
            # Going from speech to silence
            j = next((j for j in range(i + 1, len(x)) if x[j]), len(x))
            if j - i >= min_silence:
                silence = True
            else:
                x[i:j] = True
        else:
            j = i + 1
        i = j


def mh2009_vad(
    x: np.ndarray,
    sr: float,
    energy_thresh: float = 0.05,
    freq_thresh: float = 150,
    sf_thresh: float = 10,
    window: float = 0.01,
    debug: bool = False,
    min_speech: int = 5,
    min_silence: int = 10,
):
    """Trim audio clips based on a VAD algorithm adapted from [1].

    [1] M. H. Moattar and M. M. Homayounpour, "A simple but efficient
    real-time Voice Activity Detection algorithm," 2009 17th European
    Signal Processing Conference, Glasgow, 2009, pp. 2549-2553.
    https://ieeexplore.ieee.org/abstract/document/7077834
    """
    window_samples = int(sr * window)
    frame_length = 1024
    sxx = librosa.stft(
        x,
        n_fft=frame_length,
        hop_length=window_samples,
        win_length=window_samples,
        center=True,
        window="hann",
        pad_mode="reflect",
    )
    sxx = np.abs(sxx)

    # e = np.log(sxx.mean(0))
    e = librosa.feature.rms(S=sxx, frame_length=frame_length)[0]
    min_e = e.min()

    freq = librosa.fft_frequencies(sr, frame_length)[1:]
    f = freq[sxx[1:, :].argmax(0)]
    min_f = f.min()

    sf = librosa.feature.spectral_flatness(x, S=sxx, power=2)[0]
    sf = -10 * np.log10(sf / 10)
    min_sf = sf.min()

    thresh_e = energy_thresh
    thresh_f = freq_thresh
    thresh_sf = sf_thresh
    cond1 = (e - min_e >= thresh_e).astype(int)
    cond2 = (f - min_f >= thresh_f).astype(int)
    cond3 = (sf - min_sf >= thresh_sf).astype(int)
    count = cond1 + cond2 + cond3
    speech = count > 1

    filter_silence_speech(speech, min_speech=min_speech, min_silence=min_silence)

    if not np.any(speech):
        if debug:
            np.set_printoptions(precision=3, suppress=True)
            print(e - min_e)
            print(cond1)
            print(f - min_f)
            print(cond2)
            print(sf - min_sf)
            print(cond3)
        raise ValueError("No speech segments detected.")

    start = next(i for i in range(len(speech)) if speech[i])
    start *= window_samples
    end = next(i for i in range(len(speech), 0, -1) if speech[i - 1])
    end *= window_samples
    return x[start:end]


class VADTrimmer(AudioClipProcessor, fname="vad_trim", config=VADTrimmerConfig):
    config: VADTrimmerConfig

    def process_instance(self, x: np.ndarray, **kwargs) -> np.ndarray:
        sr = kwargs.pop("sr")
        if self.config.method == _VADMethod.librosa:
            return librosa.effects.trim(x, top_db=self.config.top_db)[0]
        elif self.config.method == _VADMethod.mh2009:
            return mh2009_vad(
                x,
                sr=sr,
                energy_thresh=self.config.energy_thresh,
                freq_thresh=self.config.freq_thresh,
                sf_thresh=self.config.sf_thresh,
                window=self.config.window,
                min_speech=self.config.min_speech,
                min_silence=self.config.min_silence,
            )
        raise NotImplementedError()



import os
import subprocess
import sys
from datetime import datetime
from pathlib import Path
from queue import Queue
from threading import Lock, Thread
from typing import Tuple

import click
from click_option_group import RequiredMutuallyExclusiveOptionGroup, optgroup


@click.command()
@click.argument(
    "input", type=click.Path(exists=True, dir_okay=False, path_type=Path), nargs=-1
)
@click.option("--failed", type=Path, help="Where to log failed commands.")
@optgroup.group("Parallel method", cls=RequiredMutuallyExclusiveOptionGroup)
@optgroup.option("--cpus", type=int, help="Number of CPU threads to use")
@optgroup.option("--gpus", help="GPU IDs to run on.")
def main(input: Tuple[Path], cpus: int, gpus: str, failed: Path):
    """Runs all commands specified in the INPUT file(s), splitting the
    work across multiple CPU threads or GPUs such that each command runs
    solely on whichever thread/GPU is next available.

    Each GPU still has it's own thread to run processed using
    CUDA_VISIBLE_DEVICES. Each thread reads from a synchronous queue and
    runs the command in a shell.
    """
    if len(input) == 0:
        raise ValueError("No input files specified.")

    q: "Queue[str]" = Queue()
    file_lock = Lock()
    fail_file = open(failed, "w")

    def worker(t_id: int = None, gpu: int = None):
        env = os.environ.copy()
        env.update({"PYTHONUNBUFFERED": "1"})
        if gpu is not None:
            env.update({"CUDA_VISIBLE_DEVICES": str(gpu)})
        print_prefix = f"[T{t_id}]" if gpu is None else f"[GPU{gpu}]"
        while not q.empty():
            cmd = q.get()
            print(
                datetime.now().isoformat(),
                f"Executing command on thread {t_id or gpu}: {cmd}",
                flush=True,
            )
            proc = subprocess.Popen(
                cmd,
                env=env,
                shell=True,
                text=True,
                bufsize=1,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
            )
            if proc.stdout is not None:
                with proc.stdout:
                    for line in iter(proc.stdout.readline, ""):
                        sys.stdout.write(
                            f"{datetime.now().isoformat()} {print_prefix}:{line}"
                        )
                        sys.stdout.flush()
            if proc.wait() != 0:
                with file_lock:
                    fail_file.write(cmd + "\n")
                    fail_file.flush()
            q.task_done()

    for file in input:
        with open(file) as fid:
            for line in list(filter(None, map(str.strip, fid))):
                q.put(line)
    print(f"Loaded {q.qsize()} commands")

    if cpus is not None:
        threads = [Thread(target=worker, kwargs={"t_id": i}) for i in range(cpus)]
    else:
        _gpus = list(map(int, gpus.split(",")))
        threads = [Thread(target=worker, kwargs={"gpu": i}) for i in _gpus]
    print(f"Starting threads at {datetime.now().isoformat()}")
    for t in threads:
        t.start()
    for t in threads:
        t.join()
    print(f"Finished at {datetime.now().isoformat()}")


if __name__ == "__main__":
    main()



from functools import partial
from typing import Callable, List, Union

import numpy as np
import pandas as pd
from scipy.stats import friedmanchisquare, rankdata
from sklearn.metrics.pairwise import pairwise_distances
from statsmodels.stats.libqsturng import qsturng

Matrix = List[List[float]]


def friedman_nemenyi(table: pd.DataFrame, alpha: float = 0.05):
    """Runs Friedman test on given table and optionally graphs a
    critical-difference diagram.

    Parameters
    ----------
    table: DataFrame
        The data table, with subjects as rows and independent variable
        (condition) as columns.
    alpha: float
        Significance level, must be in the range (0, 1), default is
        0.05.

    Returns
    -------
    pval: float
        The p-value for the Friedman test.
    cd: float
        The critical difference from the Nemenyi post-hoc test.
    df: pd.DataFrame
        A table containing statistics relating to ranking and average
        values of the condiions. The dataframe has these columns:
        "mean_rank", "mean", "std", "median", "mad", "effect_size".
    """
    _, pval = friedmanchisquare(*table.transpose().to_numpy())
    names = list(table.columns)
    avgrank = rankdata(-table.to_numpy(), axis=1).mean(0)
    df = pd.DataFrame(
        {
            "mean_rank": avgrank,
            "mean": table.mean(),
            "std": table.std(),
            "median": table.median(),
            "mad": table.mad(),
        },
        index=names,
    ).sort_values("mean_rank")

    topclf = df.index[0]
    n, k = table.shape
    # Effect size is calculated in terms of differences in MAD
    df["effect_size"] = (df.loc[topclf, "median"] - df["median"]) / np.sqrt(
        ((n - 1) * df.loc[topclf, "mad"] ** 2 + (n - 1) * df["mad"] ** 2) / (2 * n - 2)
    )
    cd = qsturng(1 - alpha, k, np.inf) * np.sqrt((k * (k + 1)) / (12 * n))
    return pval, cd, df


def _get_dist_func(metric: Union[Callable, str], **kwargs):
    if callable(metric):
        return partial(metric, **kwargs)
    else:
        if metric != "minkowski" and "p" in kwargs:
            del kwargs["p"]
        if metric != "mahalanobis" and "VI" in kwargs:
            del kwargs["VI"]
        return partial(pairwise_distances, metric=metric, **kwargs)


def bhattacharyya_dist(x: np.ndarray, y: np.ndarray, pinv: bool = False):
    """Calculate Bhattacharyya distance between multivariate Gaussian
    distributions.

    Parameters
    ----------
    x: array-like
        Data matrix of shape (n1_samples, n_features) corresponding to
        the first group.
    y: array-like
        Data matrix of shape (n2_samples, n_features) corresponding to
        the second group.
    pinv: bool
        Use pseudoinverse instead of inverse. This is useful if the
        covariance matrices don't have full rank or otherwise aren't
        invertible.
    """
    mu1 = np.expand_dims(np.mean(x, axis=0), 1)
    mu2 = np.expand_dims(np.mean(y, axis=0), 1)
    cov1 = np.cov(x, rowvar=False)
    cov2 = np.cov(y, rowvar=False)
    cov = (cov1 + cov2) / 2
    _, ldet1 = np.linalg.slogdet(cov1)
    _, ldet2 = np.linalg.slogdet(cov2)
    _, ldet = np.linalg.slogdet(cov)
    if pinv:
        covinv = np.linalg.pinv(cov, hermitian=True, rcond=1e-8)
    else:
        covinv = np.linalg.inv(cov)
    db = (mu1 - mu2).T @ covinv @ (mu1 - mu2) / 8 + ldet / 2 - ldet1 / 4 - ldet2 / 4

    return db.item()


def corr_ratio(x: np.ndarray, groups: Union[List[int], np.ndarray]):
    """Calculates correlation ratio for each feature using the given
    groups.

    Parameters
    ----------
    data: numpy.ndarray
        Data matrix, with shape (n_instances, n_features).
    groups: list or numpy.ndarray
        1D array of groups assignments of length n_instances. Groups
        should be labelled from 0 to G - 1 inclusive, where G is the
        number of groups.

    Returns
    -------
    eta: numpy.ndarray
        1D array of correlation coefficients of length n_features. Each
        value is in [0, 1] except if a feature takes only one value, in
        which case eta will be nan.
    """
    groups = np.array(groups)
    n_groups = groups.max() + 1
    counts = np.bincount(groups)
    mean = x.mean(0)
    g_means = np.empty((n_groups, x.shape[1]))
    for g in range(n_groups):
        g_means[g, :] = x[groups == g].mean(0)
    num = np.sum(counts[:, None] * (g_means - mean) ** 2, axis=0)
    den = np.sum((x - mean) ** 2, axis=0)
    old_err = np.seterr(divide="ignore", invalid="ignore")
    eta2 = num / den
    np.seterr(**old_err)
    return np.sqrt(eta2)


def dunn(
    x: np.ndarray,
    clusters: Union[List[int], np.ndarray],
    intra_method: str = "mean",
    inter_method: str = "cent",
    metric: Union[Callable, str] = "l2",
    p: int = 2,
):
    """Calculates the Dunn index for cluster "goodness".

    Parameters
    ----------
    data: numpy.ndarray
        Data matrix, with shape (n_instances, n_features).
    clusters: list or numpy.ndarray
        1D array of cluster assignments of length n_instances. Clusters
        should be labelled from 0 to C - 1 inclusive, where C is the
        number of clusters.
    intra_method: str
        Method for calculating intra-cluster distance. One of "max",
        "mean", "cent".
    inter_method: str
        Method for calculating inter-cluster distance. One of "cent".
    metric: str or callable
        Distance metric. If str, must be one of the sklearn or scipy
        distance methods. If callable, must take one positional argument
        and return a pairwise distance matrix.
    p: int
        Value of p for p-norm when using "lp" distance metric.

    Returns
    -------
    dunn: float
        The Dunn index for this data and cluster assignment.
    """
    clusters = np.array(clusters, dtype=int)
    n_clusters = clusters.max() + 1
    d = _get_dist_func(metric, p=p)

    intra = np.zeros(n_clusters)
    for c in range(n_clusters):
        clust_data = x[clusters == c]
        if intra_method == "max":
            idx = np.triu_indices(len(clust_data))
            intra[c] = d(clust_data)[idx].max()
        elif intra_method == "mean":
            idx = np.triu_indices(len(clust_data))
            intra[c] = d(clust_data)[idx].mean()
        elif intra_method == "cent":
            mean = clust_data.mean(0)
            intra[c] = d(clust_data, mean[None, :]).mean()

    inter = np.zeros((n_clusters, n_clusters))
    for i in range(n_clusters):
        inter[i, i] = np.inf  # To avoid min = 0
        for j in range(i + 1, n_clusters):
            if inter_method == "cent":
                mean_i = x[clusters == i].mean(0)
                mean_j = x[clusters == j].mean(0)
                inter[i, j] = inter[j, i] = d(mean_i[None, :], mean_j[None, :])

    return inter.min() / intra.max()


def kappa(data: np.ndarray):
    """Calculates Fleiss' kappa for inter-rater agreement.

    Parameters
    ----------
    data: numpy.ndarray
        The data matrix, in the form (raters x units).
    """
    cats = np.unique(data)
    n, N = data.shape

    counts = np.stack([np.sum(data == c, 0) for c in cats], 1)

    p_j = np.sum(counts, axis=0) / (N * n)
    assert np.isclose(np.sum(p_j), 1)
    Pe = np.sum(p_j ** 2)

    P = (np.sum(counts ** 2, 1) - n) / (n * (n - 1))
    Pbar = np.mean(P)

    return (Pbar - Pe) / (1 - Pe)


class Deltas:
    @staticmethod
    def nominal(c: int, k: int):
        return float(c != k)

    @staticmethod
    def interval(c: float, k: float):
        return (c - k) ** 2


def alpha(
    data: np.ndarray,
    delta: Union[Callable[[int, int], float], List[List[float]], str] = "nominal",
):
    """Calculates Krippendorff's alpha coefficient [1, sec. 11.3] for
    inter-rater agreement.

    [1] K. Krippendorff, Content analysis: An introduction to its
    methodology. Sage publications, 2004.

    Parameters
    ----------
    data: numpy.ndarray
        The data matrix, shape (n_raters, n_units). Each cell (i, j)
        represents the value assigned to unit j by rater i, or 0
        representing no response.
    delta: callable, 2-D array-like or str
        The delta metric. Default is the nominal metric, which takes the
        value 1 in case c != k and 0 otherwise.
    """
    # The following implementation was based off the Wikipedia article:
    # https://en.wikipedia.org/wiki/Krippendorff%27s_alpha

    # Response categories go from 1 to R, 0 represents no response
    R = np.max(data)
    counts = np.apply_along_axis(lambda x: np.bincount(x, minlength=R + 1), 0, data).T
    count_sum = np.sum(counts, 0)
    assert len(count_sum) == R + 1

    def ordinal(c: int, k: int):
        if k < c:
            c, k = k, c
        s = (
            sum(count_sum[g] for g in range(c, k + 1))
            - (count_sum[c] + count_sum[k]) / 2
        )
        return s ** 2

    if isinstance(delta, str):
        delta = {
            "nominal": Deltas.nominal,
            "ordinal": ordinal,
            "interval": Deltas.interval,
        }[delta]

    if not callable(delta):
        try:
            delta[0][0]
        except IndexError:
            raise TypeError("delta must be either str, callable or 2D array.")

        def _delta(c, k):
            new_delta = delta
            return new_delta[c][k]

        delta = _delta

    m_u = np.sum(counts[:, 1:], 1)

    valid = m_u >= 2
    counts = counts[valid]
    m_u = m_u[valid]
    data = data[:, valid]

    n = np.sum(m_u)

    n_cku = np.matmul(counts[:, :, None], counts[:, None, :])
    for i in range(R + 1):
        n_cku[:, i, i] = counts[:, i] * (counts[:, i] - 1)

    D_o = 0
    for c in range(1, R + 1):
        for k in range(1, R + 1):
            D_o += delta(c, k) * n_cku[:, c, k]
    D_o = np.sum(D_o / (n * (m_u - 1)))

    D_e = 0
    P_ck = np.bincount(data.flat)
    for c in range(1, R + 1):
        for k in range(1, R + 1):
            D_e += delta(c, k) * P_ck[c] * P_ck[k]
    D_e /= n * (n - 1)

    return 1 - D_o / D_e



import logging
from collections import defaultdict
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import tensorflow as tf
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.model_selection import BaseCrossValidator
from sklearn.pipeline import Pipeline
from tensorflow.keras.callbacks import History, TensorBoard
from tensorflow.keras.metrics import SparseCategoricalAccuracy

from ertk.tensorflow.utils import (
    DataFunction,
    TFModelFunction,
    init_gpu_memory_growth,
    tf_dataset_gen,
    tf_dataset_mem,
    tf_dataset_mem_ragged,
)
from ertk.train import get_scores
from ertk.utils import ScoreFunction, filter_kwargs


def tf_train_val_test(
    model_fn: TFModelFunction,
    train_data: Tuple[np.ndarray, ...],
    valid_data: Tuple[np.ndarray, ...],
    test_data: Optional[Tuple[np.ndarray, ...]] = None,
    scoring: Union[
        str, List[str], Dict[str, ScoreFunction], Callable[..., float]
    ] = "accuracy",
    data_fn: Union[DataFunction, str] = None,
    batch_size: int = 32,
    verbose: int = 0,
    fit_params: Dict[str, Any] = {},
) -> Dict[str, Union[float, History]]:
    """Trains on given data, using given validation data, and tests on
    given test data.

    Parameters
    ----------
    model_fn: callable
        The function used to create a compiled Model.
    train_data: tuple
        The training data. Tuple elements will be used as positional
        arguments to data_fn.
    valid_data: tuple
        Validation data.
    test_data: tuple, optional
        Optional test data. If not given, will use validation data as
        test data.
    scoring: str, or callable, or list of str, or dict of str to callable
        The scoring to use. Same requirements as for sklearn
        cross_validate().
    data_fn: str or callable
        A string, or callable that returns a tensorflow.data.Dataset
        instance which yields data batches. The callable should take x,
        y and sample_weight parameters, as well as shuffle and
        batch_size keyword arguments. Valid string values are: {"mem",
        "gen", "mem_ragged"}. "mem" indicates an in-memory Dataset.
        "gen" indicates a generator dataset, and "ragged" indicates
        taking variable-length input.
    fit_params: dict, optional
        Any keyword arguments to supply to the Keras fit() method.
        Default is no keyword arguments.

    Returns
    -------
    scores: dict
        A dictionary with scorer names as keys and scores as values.
    """
    init_gpu_memory_growth()

    if test_data is None:
        test_data = valid_data

    data_fn = fit_params.pop("data_fn", data_fn)
    batch_size = fit_params.pop("batch_size", batch_size)
    if "sample_weight" in fit_params:
        del fit_params["sample_weight"]  # Part of *_data

    data_fns = {
        "mem": tf_dataset_mem,
        "gen": tf_dataset_gen,
        "mem_ragged": tf_dataset_mem_ragged,
    }
    if data_fn is None:
        if len(train_data[0].shape) == 3:
            data_fn = tf_dataset_gen
        elif len(train_data[0].shape) == 1:
            data_fn = tf_dataset_mem_ragged
        else:
            data_fn = tf_dataset_mem
    elif isinstance(data_fn, str):
        data_fn = data_fns[data_fn]
    elif not callable(data_fn):
        raise ValueError(f"Unsupported value for data_fn {data_fn}")

    logging.debug(f"fit_params={fit_params}")
    logging.debug(f"batch_size={batch_size}")
    logging.debug(f"data_fn={data_fn}")

    tf.keras.backend.clear_session()

    clf = model_fn()
    logging.debug(clf)
    if isinstance(clf, Pipeline):
        new_pipeline = clf[:-1]
        clf = clf._final_estimator
        new_pipeline.fit(train_data[0], y=train_data[1])
        train_data = (new_pipeline.transform(train_data[0]), *train_data[1:])
        valid_data = (new_pipeline.transform(valid_data[0]), *valid_data[1:])
        test_data = (new_pipeline.transform(test_data[0]), *test_data[1:])
    if not isinstance(clf, tf.keras.models.Model):
        raise TypeError(
            "model_fn must return a Model or a Pipeline which ends in a Model."
        )

    train_dataset = data_fn(*train_data, batch_size=batch_size)
    valid_dataset = data_fn(*valid_data, batch_size=batch_size, shuffle=False)
    test_dataset = data_fn(*test_data, batch_size=batch_size, shuffle=False)

    clf.summary(print_fn=logging.info)

    fit_params = filter_kwargs(fit_params, clf.fit)
    history = clf.fit(
        train_dataset, validation_data=valid_dataset, verbose=verbose, **fit_params
    )
    # Setting y_true is necessary for dataset creation routines that may
    # reorder data on creation (but without shuffling each time).
    y_true = np.concatenate([x[1] for x in test_dataset.as_numpy_iterator()])
    y_pred = tf.argmax(clf.predict(test_dataset), axis=-1)

    scores = get_scores(scoring, y_pred, y_true)
    scores = {f"test_{k}": v for k, v in scores.items()}
    scores["history"] = history
    return scores


def tf_cross_validate(
    model_fn: TFModelFunction,
    x: np.ndarray,
    y: np.ndarray,
    *,
    cv: BaseCrossValidator,
    groups: Optional[np.ndarray] = None,
    verbose: int = 0,
    scoring: Union[
        str, List[str], Dict[str, ScoreFunction], Callable[..., float]
    ] = "accuracy",
    fit_params: Dict[str, Any] = {},
):
    """Performs cross-validation on a TensorFlow model. This works with
    both sequence models and single vector models.

    Parameters
    ----------
    model_fn: callable,
        The function used to create a compiled Keras model. This is
        called repeatedly on each iteration of cross-validation.
    x: numpy.ndarray,
        The data array. For sequence input this will be a (ragged) 3-D
        array (array of arrays). Otherwise it will be a contiguous 2-D
        matrix.
    y: numpy.ndarray,
        A 1-D array of shape (n_instances,) containing the data labels.
    groups: np.ndarray, optional
        The groups to use for some cross-validation splitters (e.g.
        LeaveOneGroupOut).
    cv: BaseCrossValidator,
        The CV split generator to use.
    scoring: str, or callable, or list of str, or dict of str to callable
        The scoring to use. Same requirements as for sklearn
        cross_validate().
    data_fn: callable
        A callable that returns a tensorflow.data.Dataset instance which
        yields data batches. The call signature of data_fn should be
        data_fn(x, y, shuffle=True, **kwargs).
    fit_params: dict, optional
        Any keyword arguments to supply to the Keras fit() method.
        Default is no keyword arguments.
    """
    init_gpu_memory_growth()

    log_dir = fit_params.pop("log_dir", None)
    sw = fit_params.pop("sample_weight", None)
    data_fn = fit_params.pop("data_fn", None)

    logging.debug(f"log_dir={log_dir}")
    logging.debug(f"sample_weight={sw}")
    logging.debug(f"data_fn={data_fn}")
    logging.debug(f"cv={cv}")
    logging.debug(f"fit_params={fit_params}")

    n_folds = cv.get_n_splits(x, y, groups)
    scores = defaultdict(list)
    for fold, (train, test) in enumerate(cv.split(x, y, groups)):
        if verbose:
            logging.info(f"Fold {fold + 1}/{n_folds}")

        x_train = x[train]
        y_train = y[train]
        x_test = x[test]
        y_test = y[test]
        sw_train = sw[train] if sw is not None else None
        sw_test = sw[test] if sw is not None else None

        callbacks = []
        if log_dir is not None:
            callbacks.append(
                TensorBoard(
                    log_dir=log_dir / str(fold + 1),
                    profile_batch=0,
                    write_graph=False,
                    write_images=False,
                )
            )

        fit_params["callbacks"] = callbacks
        _scores = tf_train_val_test(
            model_fn,
            train_data=(x_train, y_train, sw_train),
            valid_data=(x_test, y_test, sw_test),
            test_data=(x_test, y_test, sw_test),
            data_fn=data_fn,
            scoring=scoring,
            verbose=verbose,
            fit_params=dict(fit_params),
        )

        for k in _scores:
            scores[k].append(_scores[k])
    return {k: np.array(scores[k]) for k in scores}


class TFClassifierWrapper(ClassifierMixin, BaseEstimator):
    """Class wrapper for a TensorFlow Keras classifier model.

    Parameters
    ----------
    model_fn: callable
        A callable that returns a compiled Model that can be trained.
    n_epochs: int, optional, default = 50
        Maximum number of epochs to train for.
    class_weight: dict, optional
        A dictionary mapping class IDs to weights. Default is to ignore
        class weights.
    data_fn: callable, optional
        Callable that takes x and y as input and returns a
        tensorflow.keras.Sequence object or a tensorflow.data.Dataset
        object.
    callbacks: list, optional
        A list of tensorflow.keras.callbacks.Callback objects to use during
        training. Default is an empty list, so that the default Keras
        callbacks are used.
    verbose: bool, default = False
        Whether to output details per epoch.
    """

    def __init__(
        self,
        model_fn: TFModelFunction,
        data_fn: Optional[DataFunction] = None,
        fit_params: Dict = {},
        val_method: Optional[str] = None,
    ):
        self.model_fn = model_fn
        self._data_fn = data_fn
        self.fit_params = fit_params.copy()
        self.val_method = val_method

    def data_fn(
        self,
        x: np.ndarray,
        y: np.ndarray,
        sw: Optional[np.ndarray] = None,
        *,
        shuffle: bool = True,
    ) -> tf.data.Dataset:
        if self._data_fn is not None:
            return self._data_fn(x, y, sw, shuffle=shuffle)
        if sw is not None:
            dataset = tf.data.Dataset.from_tensor_slices((x, y, sw))
        else:
            dataset = tf.data.Dataset.from_tensor_slices((x, y))
        if shuffle:
            dataset = dataset.shuffle(len(x))
        return dataset

    def fit(
        self,
        x: np.ndarray,
        y: np.ndarray,
        sample_weight: np.ndarray = None,
    ):
        # Clear graph
        tf.keras.backend.clear_session()
        self.model = self.model_fn()

        train_data = self.data_fn(x, y, sample_weight, shuffle=True)
        if self.val_method is not None:
            valid_data = train_data
        else:
            valid_data = None
        return self.model.fit(
            train_data,
            validation_data=valid_data,
            **self.fit_params,
        )

    def predict(self, x: np.ndarray) -> np.ndarray:
        test_data = self.data_fn(x, np.zeros(len(x)), shuffle=False)
        y_pred = np.argmax(self.model.predict(test_data), axis=-1)
        return y_pred


class BalancedSparseCategoricalAccuracy(SparseCategoricalAccuracy):
    """Calculates categorical accuracy with class weights inversely
    proportional to their size. This behaves as if classes are balanced
    having the same number of instances, and is equivalent to the
    arithmetic mean recall over all classes.
    """

    def __init__(self, name="balanced_sparse_categorical_accuracy", **kwargs):
        super().__init__(name, **kwargs)

    def update_state(self, y_true, y_pred, sample_weight=None):
        y_flat = y_true
        if y_true.shape.ndims == y_pred.shape.ndims:
            y_flat = tf.squeeze(y_flat, axis=[-1])
        y_true_int = tf.cast(y_flat, tf.int32)

        cls_counts = tf.math.bincount(y_true_int)
        cls_counts = tf.math.reciprocal_no_nan(tf.cast(cls_counts, self.dtype))
        weight = tf.gather(cls_counts, y_true_int)
        return super().update_state(y_true, y_pred, sample_weight=weight)


def tf_classification_metrics():
    return [
        SparseCategoricalAccuracy(name="war"),
        BalancedSparseCategoricalAccuracy(name="uar"),
    ]
